"""Defines the preprocessing configurations that define the ensembling of
different members.
"""

#  Copyright (c) Prior Labs GmbH 2025.

from __future__ import annotations

import warnings
from collections.abc import Callable, Iterable, Iterator, Sequence
from dataclasses import dataclass, field
from functools import partial
from itertools import chain, product, repeat
from typing import TYPE_CHECKING, Literal, TypeVar
from typing_extensions import override

import joblib
import numpy as np
import torch
from torch.utils.data import Dataset

from tabpfn.architectures.base.bar_distribution import FullSupportBarDistribution
from tabpfn.constants import (
    CLASS_SHUFFLE_OVERESTIMATE_FACTOR,
    MAXIMUM_FEATURE_SHIFT,
    PARALLEL_MODE_TO_RETURN_AS,
    SUPPORTS_RETURN_AS,
)
from tabpfn.preprocessors import (
    AddFingerprintFeaturesStep,
    DifferentiableZNormStep,
    EncodeCategoricalFeaturesStep,
    FeaturePreprocessingTransformerStep,
    NanHandlingPolynomialFeaturesStep,
    RemoveConstantFeaturesStep,
    ReshapeFeatureDistributionsStep,
    SequentialFeatureTransformer,
    ShuffleFeaturesStep,
)
from tabpfn.utils import infer_random_state

if TYPE_CHECKING:
    import numpy.typing as npt
    from sklearn.base import TransformerMixin
    from sklearn.pipeline import Pipeline


T = TypeVar("T")


def balance(x: Iterable[T], n: int) -> list[T]:
    """Take a list of elements and make a new list where each appears `n` times.

    E.g. balance([1, 2, 3], 2) -> [1, 1, 2, 2, 3, 3]
    """
    return list(chain.from_iterable(repeat(elem, n) for elem in x))


@dataclass
class BaseDatasetConfig:
    """Base configuration class for holding dataset specifics."""

    config: EnsembleConfig
    X_raw: np.ndarray | torch.Tensor
    y_raw: np.ndarray | torch.Tensor
    cat_ix: list[int]


@dataclass
class ClassifierDatasetConfig(BaseDatasetConfig):
    """Classification Dataset + Model Configuration class."""


@dataclass
class RegressorDatasetConfig(BaseDatasetConfig):
    """Regression Dataset + Model Configuration class."""

    znorm_space_bardist_: FullSupportBarDistribution | None = field(default=None)

    @property
    def bardist_(self) -> FullSupportBarDistribution:
        """DEPRECATED: Accessing `bardist_` is deprecated.
        Use `znorm_space_bardist_` instead.
        """
        warnings.warn(
            "`bardist_` is deprecated and will be removed in a future version. "
            "Please use `znorm_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.znorm_space_bardist_

    @bardist_.setter
    def bardist_(self, value: FullSupportBarDistribution) -> None:
        """DEPRECATED: Setting `bardist_` is deprecated.
        Use `znorm_space_bardist_`.
        """
        warnings.warn(
            "`bardist_` is deprecated and will be removed in a future version. "
            "Please use `znorm_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.znorm_space_bardist_ = value


@dataclass(frozen=True, eq=True)
class PreprocessorConfig:
    """Configuration for data preprocessors.

    Attributes:
        name: Name of the preprocessor.
        categorical_name:
            Name of the categorical encoding method.
            Options: "none", "numeric", "onehot", "ordinal", "ordinal_shuffled", "none".
        append_to_original: If set to "auto", this is dynamically set to
            True if the number of features is less than 500, and False otherwise.
            Note that if set to "auto" and `max_features_per_estimator` is set as well,
            this flag will become False if the number of features is larger than
            `max_features_per_estimator / 2`. If True, the transformed features are
            appended to the original features, however both are capped at the
            max_features_per_estimator threshold, this should be used with caution as a
            given model might not be configured for it.
        max_features_per_estimator: Maximum number of features per estimator. In case
            the dataset has more features than this, the features are subsampled for
            each estimator independently. If append to original is set to True we can
            still have more features.
        global_transformer_name: Name of the global transformer to use.
    """

    name: Literal[
        "per_feature",  # a different transformation for each feature
        "power",  # a standard sklearn power transformer
        "safepower",  # a power transformer that prevents some numerical issues
        "power_box",
        "safepower_box",
        "quantile_uni_coarse",  # quantile transformations with few quantiles up to many
        "quantile_norm_coarse",
        "quantile_uni",
        "quantile_norm",
        "quantile_uni_fine",
        "quantile_norm_fine",
        "squashing_scaler_default",
        "squashing_scaler_max10",
        "robust",  # a standard sklearn robust scaler
        "kdi",
        "none",  # no transformation (only standardization in transformer)
        "kdi_random_alpha",
        "kdi_uni",
        "kdi_random_alpha_uni",
        "adaptive",
        "norm_and_kdi",
        # KDI with alpha collection
        "kdi_alpha_0.3_uni",
        "kdi_alpha_0.5_uni",
        "kdi_alpha_0.8_uni",
        "kdi_alpha_1.0_uni",
        "kdi_alpha_1.2_uni",
        "kdi_alpha_1.5_uni",
        "kdi_alpha_2.0_uni",
        "kdi_alpha_3.0_uni",
        "kdi_alpha_5.0_uni",
        "kdi_alpha_0.3",
        "kdi_alpha_0.5",
        "kdi_alpha_0.8",
        "kdi_alpha_1.0",
        "kdi_alpha_1.2",
        "kdi_alpha_1.5",
        "kdi_alpha_2.0",
        "kdi_alpha_3.0",
        "kdi_alpha_5.0",
    ]
    categorical_name: Literal[
        # categorical features are pretty much treated as ordinal, just not resorted
        "none",
        # categorical features are treated as numeric,
        # that means they are also power transformed for example
        "numeric",
        # "onehot": categorical features are onehot encoded
        "onehot",
        # "ordinal": categorical features are sorted and encoded as
        # integers from 0 to n_categories - 1
        "ordinal",
        # "ordinal_shuffled": categorical features are encoded as integers
        # from 0 to n_categories - 1 in a random order
        "ordinal_shuffled",
        "ordinal_very_common_categories_shuffled",
    ] = "none"
    append_original: bool | Literal["auto"] = False
    max_features_per_estimator: int = 500
    global_transformer_name: (
        Literal[
            "scaler",
            "svd",
            "svd_quarter_components",
        ]
        | None
    ) = None
    differentiable: bool = False

    @override
    def __str__(self) -> str:
        return (
            f"{self.name}_cat:{self.categorical_name}"
            + ("_and_none" if self.append_original else "")
            + (f"_max_feats_per_est_{self.max_features_per_estimator}")
            + (
                f"_global_transformer_{self.global_transformer_name}"
                if self.global_transformer_name is not None
                else ""
            )
        )


def default_classifier_preprocessor_configs() -> list[PreprocessorConfig]:
    """Get default preprocessor configurations for classification.

    These are the defaults used when training new models, which will then be stored in
    the model checkpoint.

    See `v2_classifier_preprocessor_configs()`, `v2_5_classifier_preprocessor_configs()`
    for the preprocessing used earlier versions of the model.
    """
    return [
        PreprocessorConfig(
            name="squashing_scaler_default",
            append_original=False,
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd_quarter_components",
            max_features_per_estimator=500,
        ),
        PreprocessorConfig(
            name="none",
            categorical_name="numeric",
            max_features_per_estimator=500,
        ),
    ]


def default_regressor_preprocessor_configs() -> list[PreprocessorConfig]:
    """Default preprocessor configurations for regression.

    These are the defaults used when training new models, which will then be stored in
    the model checkpoint.

    See `v2_regressor_preprocessor_configs()`, `v2_5_regressor_preprocessor_configs()`
    for the preprocessing used earlier versions of the model.
    """
    return [
        PreprocessorConfig(
            name="quantile_uni_coarse",
            append_original="auto",
            categorical_name="numeric",
            global_transformer_name=None,
            max_features_per_estimator=500,
        ),
        PreprocessorConfig(
            name="squashing_scaler_default",
            append_original=False,
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd_quarter_components",
            max_features_per_estimator=500,
        ),
    ]


# Feature subsampling was disabled in v2, so choose a threshold that will never be
# reached.
_V2_FEATURE_SUBSAMPLING_THRESHOLD = 1_000_000


def v2_classifier_preprocessor_configs() -> list[PreprocessorConfig]:
    """Get the preprocessor configuration for classification in v2 of the model."""
    return [
        PreprocessorConfig(
            "quantile_uni_coarse",
            append_original="auto",
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd",
            max_features_per_estimator=_V2_FEATURE_SUBSAMPLING_THRESHOLD,
        ),
        PreprocessorConfig(
            "none",
            categorical_name="numeric",
            max_features_per_estimator=_V2_FEATURE_SUBSAMPLING_THRESHOLD,
        ),
    ]


def v2_regressor_preprocessor_configs() -> list[PreprocessorConfig]:
    """Get the preprocessor configuration for regression in v2 of the model."""
    return [
        PreprocessorConfig(
            "quantile_uni",
            append_original=True,
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd",
        ),
        PreprocessorConfig("safepower", categorical_name="onehot"),
    ]


def v2_5_classifier_preprocessor_configs() -> list[PreprocessorConfig]:
    """Get the preprocessor configuration for classification in v2.5 of the model."""
    return [
        PreprocessorConfig(
            name="squashing_scaler_default",
            append_original=False,
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd_quarter_components",
            max_features_per_estimator=500,
        ),
        PreprocessorConfig(
            name="none",
            categorical_name="numeric",
            max_features_per_estimator=500,
        ),
    ]


def v2_5_regressor_preprocessor_configs() -> list[PreprocessorConfig]:
    """Get the preprocessor configuration for regression in v2.5 of the model."""
    return [
        PreprocessorConfig(
            name="quantile_uni_coarse",
            append_original="auto",
            categorical_name="numeric",
            global_transformer_name=None,
            max_features_per_estimator=500,
        ),
        PreprocessorConfig(
            name="squashing_scaler_default",
            append_original=False,
            categorical_name="ordinal_very_common_categories_shuffled",
            global_transformer_name="svd_quarter_components",
            max_features_per_estimator=500,
        ),
    ]


def generate_index_permutations(
    n: int,
    *,
    max_index: int,
    subsample: int | float,
    random_state: int | np.random.Generator | None,
) -> list[npt.NDArray[np.int64]]:
    """Generate indices for subsampling from the data.

    Args:
        n: Number of indices to generate.
        max_index: Maximum index to generate.
        subsample:
            Number of indices to subsample. If `int`, subsample that many
            indices. If float, subsample that fraction of indices.
        random_state: Random number generator.

    Returns:
        List of indices to subsample.
    """
    _, rng = infer_random_state(random_state)
    if isinstance(subsample, int):
        if subsample < 1:
            raise ValueError(f"{subsample=} must be larger than 1 if int")
        subsample = min(subsample, max_index)

        return [rng.permutation(max_index)[:subsample] for _ in range(n)]

    if isinstance(subsample, float):
        if not (0 < subsample < 1):
            raise ValueError(f"{subsample=} must be in (0, 1) if float")
        subsample = int(subsample * max_index) + 1
        return [rng.permutation(max_index)[:subsample] for _ in range(n)]

    raise ValueError(f"{subsample=} must be int or float.")


# TODO: (Klemens)
# Make this frozen (frozen=True)
@dataclass
class EnsembleConfig:
    """Configuration for an ensemble member.

    Attributes:
        preprocess_config: Preprocessor configuration to use.
        add_fingerprint_feature: Whether to add fingerprint features.
        polynomial_features: Maximum number of polynomial features to add, if any.
        feature_shift_count: How much to shift the features columns.
        feature_shift_decoder: How to shift features.
        subsample_ix: Indices of samples to use for this ensemble member.
            If `None`, no subsampling is done.
    """

    preprocess_config: PreprocessorConfig
    add_fingerprint_feature: bool
    polynomial_features: Literal["no", "all"] | int
    feature_shift_count: int
    feature_shift_decoder: Literal["shuffle", "rotate"] | None
    subsample_ix: npt.NDArray[np.int64] | None  # OPTIM: Could use uintp
    # Internal index specifying which model to use for this ensemble member.
    _model_index: int

    @classmethod
    def generate_for_classification(  # noqa: PLR0913
        cls,
        *,
        num_estimators: int,
        subsample_size: int | float | None,
        max_index: int,
        add_fingerprint_feature: bool,
        polynomial_features: Literal["no", "all"] | int,
        feature_shift_decoder: Literal["shuffle", "rotate"] | None,
        preprocessor_configs: Sequence[PreprocessorConfig],
        class_shift_method: Literal["rotate", "shuffle"] | None,
        n_classes: int,
        random_state: int | np.random.Generator | None,
        num_models: int,
    ) -> list[ClassifierEnsembleConfig]:
        """Generate ensemble configurations for classification.

        Args:
            num_estimators: Number of ensemble configurations to generate.
            subsample_size:
                Number of samples to subsample. If int, subsample that many
                samples. If float, subsample that fraction of samples. If `None`, no
                subsampling is done.
            max_index: Maximum index to generate for.
            add_fingerprint_feature: Whether to add fingerprint features.
            polynomial_features: Maximum number of polynomial features to add, if any.
            feature_shift_decoder: How shift features
            preprocessor_configs: Preprocessor configurations to use on the data.
            class_shift_method: How to shift classes for classpermutation.
            n_classes: Number of classes.
            random_state: Random number generator.
            num_models: Number of models to use.

        Returns:
            List of ensemble configurations.
        """
        static_seed, rng = infer_random_state(random_state)
        start = rng.integers(0, MAXIMUM_FEATURE_SHIFT)
        featshifts = np.arange(start, start + num_estimators)
        featshifts = rng.choice(featshifts, size=num_estimators, replace=False)  # type: ignore

        if class_shift_method == "rotate":
            arange = np.arange(0, n_classes)
            shifts = rng.permutation(n_classes).tolist()
            class_permutations = [np.roll(arange, s) for s in shifts]
            class_permutations = [  # type: ignore
                class_permutations[c] for c in rng.choice(n_classes, num_estimators)
            ]
        elif class_shift_method == "shuffle":
            noise = rng.random(
                (num_estimators * CLASS_SHUFFLE_OVERESTIMATE_FACTOR, n_classes)
            )
            shufflings = np.argsort(noise, axis=1)
            uniqs = np.unique(shufflings, axis=0)
            balance_count = num_estimators // len(uniqs)
            class_permutations = balance(uniqs, balance_count)
            rand_count = num_estimators % len(uniqs)
            if rand_count > 0:
                class_permutations += [  # type: ignore
                    uniqs[i] for i in rng.choice(len(uniqs), size=rand_count)
                ]
        elif class_shift_method is None:
            class_permutations = [None] * num_estimators  # type: ignore
        else:
            raise ValueError(f"Unknown {class_shift_method=}")

        subsamples: list[None] | list[np.ndarray]
        if isinstance(subsample_size, (int, float)):
            subsamples = generate_index_permutations(
                n=num_estimators,
                max_index=max_index,
                subsample=subsample_size,
                random_state=static_seed,
            )
        elif subsample_size is None:
            subsamples = [None] * num_estimators  # type: ignore
        else:
            raise ValueError(
                f"Invalid subsample_samples: {subsample_size}",
            )

        balance_count = num_estimators // len(preprocessor_configs)

        # Replicate each config balance_count times
        configs_ = balance(preprocessor_configs, balance_count)
        # Number still needed to reach n
        leftover = num_estimators - len(configs_)
        if leftover > 0:
            # the preprocessor configs should be ordered by performance
            configs_.extend(preprocessor_configs[:leftover])

        # Models are simply cycled through for the estimators.
        # This ensures that different preprocessings are applied to different models.
        model_indices = [i % num_models for i in range(num_estimators)]

        return [
            ClassifierEnsembleConfig(
                preprocess_config=preprocesses_config,
                feature_shift_count=featshift,
                add_fingerprint_feature=add_fingerprint_feature,
                polynomial_features=polynomial_features,
                feature_shift_decoder=feature_shift_decoder,
                subsample_ix=subsample_ix,
                class_permutation=class_perm,
                _model_index=model_index,
            )
            for (
                featshift,
                preprocesses_config,
                subsample_ix,
                class_perm,
                model_index,
            ) in zip(
                featshifts,
                configs_,
                subsamples,
                class_permutations,
                model_indices,
            )
        ]

    @classmethod
    def generate_for_regression(
        cls,
        *,
        num_estimators: int,
        subsample_size: int | float | None,
        max_index: int,
        add_fingerprint_feature: bool,
        polynomial_features: Literal["no", "all"] | int,
        feature_shift_decoder: Literal["shuffle", "rotate"] | None,
        preprocessor_configs: Sequence[PreprocessorConfig],
        target_transforms: Sequence[TransformerMixin | Pipeline | None],
        random_state: int | np.random.Generator | None,
        num_models: int,
    ) -> list[RegressorEnsembleConfig]:
        """Generate ensemble configurations for regression.

        Args:
            num_estimators: Number of ensemble configurations to generate.
            subsample_size:
                Number of samples to subsample. If int, subsample that many
                samples. If float, subsample that fraction of samples. If `None`, no
                subsampling is done.
            max_index: Maximum index to generate for.
            add_fingerprint_feature: Whether to add fingerprint features.
            polynomial_features: Maximum number of polynomial features to add, if any.
            feature_shift_decoder: How shift features
            preprocessor_configs: Preprocessor configurations to use on the data.
            target_transforms: Target transformations to apply.
            random_state: Random number generator.
            num_models: Number of models to use.

        Returns:
            List of ensemble configurations.
        """
        static_seed, rng = infer_random_state(random_state)
        start = rng.integers(0, MAXIMUM_FEATURE_SHIFT)
        featshifts = np.arange(start, start + num_estimators)
        featshifts = rng.choice(featshifts, size=num_estimators, replace=False)  # type: ignore

        subsamples: list[None] | list[np.ndarray]
        if isinstance(subsample_size, (int, float)):
            subsamples = generate_index_permutations(
                n=num_estimators,
                max_index=max_index,
                subsample=subsample_size,
                random_state=static_seed,
            )
        elif subsample_size is None:
            subsamples = [None] * num_estimators
        else:
            raise ValueError(
                f"Invalid subsample_samples: {subsample_size}",
            )

        # Get equal representation of all preprocessor configs
        combos = list(product(preprocessor_configs, target_transforms))
        balance_count = num_estimators // len(combos)
        configs_ = balance(combos, balance_count)
        # Number still needed to reach n
        leftover = num_estimators - len(configs_)
        if leftover > 0:
            # the preprocessor configs should be ordered by performance
            # TODO: what about the target transforms?
            configs_ += combos[:leftover]

        # Models are simply cycled through for the estimators.
        # This ensures that different preprocessings and target transformations are
        # applied to different models.
        model_indices = [i % num_models for i in range(num_estimators)]

        return [
            RegressorEnsembleConfig(
                preprocess_config=preprocess_config,
                feature_shift_count=featshift,
                add_fingerprint_feature=add_fingerprint_feature,
                polynomial_features=polynomial_features,
                feature_shift_decoder=feature_shift_decoder,
                subsample_ix=subsample_ix,
                target_transform=target_transform,
                _model_index=model_index,
            )
            for featshift, subsample_ix, (
                preprocess_config,
                target_transform,
            ), model_index in zip(
                featshifts,
                subsamples,
                configs_,
                model_indices,
            )
        ]

    # TODO(eddiebergman): Make this sklearn pipeline
    def to_pipeline(
        self,
        *,
        random_state: int | np.random.Generator | None,
    ) -> SequentialFeatureTransformer:
        """Convert the ensemble configuration to a preprocessing pipeline."""
        steps: list[FeaturePreprocessingTransformerStep] = []

        if isinstance(self.polynomial_features, int):
            assert self.polynomial_features > 0, "Poly. features to add must be >0!"
            use_poly_features = True
            max_poly_features = self.polynomial_features
        elif self.polynomial_features == "all":
            use_poly_features = True
            max_poly_features = None
        elif self.polynomial_features == "no":
            use_poly_features = False
            max_poly_features = None
        else:
            raise ValueError(
                f"Invalid polynomial_features value: {self.polynomial_features}",
            )
        if use_poly_features:
            steps.append(
                NanHandlingPolynomialFeaturesStep(
                    max_features=max_poly_features,
                    random_state=random_state,
                ),
            )

        steps.append(RemoveConstantFeaturesStep())

        if self.preprocess_config.differentiable:
            steps.append(DifferentiableZNormStep())
        else:
            steps.extend(
                [
                    ReshapeFeatureDistributionsStep(
                        transform_name=self.preprocess_config.name,
                        append_to_original=self.preprocess_config.append_original,
                        max_features_per_estimator=self.preprocess_config.max_features_per_estimator,
                        global_transformer_name=self.preprocess_config.global_transformer_name,
                        apply_to_categorical=(
                            self.preprocess_config.categorical_name == "numeric"
                        ),
                        random_state=random_state,
                    ),
                    EncodeCategoricalFeaturesStep(
                        self.preprocess_config.categorical_name,
                        random_state=random_state,
                    ),
                ],
            )

        if self.add_fingerprint_feature:
            steps.append(AddFingerprintFeaturesStep(random_state=random_state))

        steps.append(
            ShuffleFeaturesStep(
                shuffle_method=self.feature_shift_decoder,
                shuffle_index=self.feature_shift_count,
                random_state=random_state,
            ),
        )
        return SequentialFeatureTransformer(steps)


@dataclass
class ClassifierEnsembleConfig(EnsembleConfig):
    """Configuration for a classifier ensemble member.

    Attributes:
        class_permutation: Permutation to apply to classes

    See [EnsembleConfig][tabpfn.preprocessing.EnsembleConfig] for more details.
    """

    class_permutation: np.ndarray | None


@dataclass
class RegressorEnsembleConfig(EnsembleConfig):
    """Configuration for a regression ensemble member.

    See [EnsembleConfig][tabpfn.preprocessing.EnsembleConfig] for more details.
    """

    target_transform: TransformerMixin | Pipeline | None


def fit_preprocessing_one(
    config: EnsembleConfig,
    X_train: np.ndarray | torch.Tensor,
    y_train: np.ndarray | torch.Tensor,
    random_state: int | np.random.Generator | None = None,
    *,
    cat_ix: list[int],
) -> tuple[
    EnsembleConfig,
    SequentialFeatureTransformer,
    np.ndarray,
    np.ndarray,
    list[int],
]:
    """Fit preprocessing pipeline for a single ensemble configuration.

    Args:
        config: Ensemble configuration.
        X_train: Training data.
        y_train: Training target.
        random_state: Random seed.
        cat_ix: Indices of categorical features.
        process_idx: Which indices to consider. Only return values for these indices.
            if None, all indices are processed, which is the default.

    Returns:
        Tuple containing the ensemble configuration, the fitted preprocessing pipeline,
        the transformed training data, the transformed target, and the indices of
        categorical features.
    """
    static_seed, _ = infer_random_state(random_state)
    if config.subsample_ix is not None:
        X_train = X_train[config.subsample_ix]
        y_train = y_train[config.subsample_ix]
    if not isinstance(X_train, torch.Tensor):
        X_train = X_train.copy()
        y_train = y_train.copy()

    preprocessor = config.to_pipeline(random_state=static_seed)
    res = preprocessor.fit_transform(X_train, cat_ix)

    # TODO(eddiebergman): Not a fan of this, wish it was more transparent, but we want
    # to distuinguish what to do with the `ys` based on the ensemble config type

    # TODO: (Klemens)
    y_train_processed = transform_labels_one(config, y_train)

    return (config, preprocessor, res.X, y_train_processed, res.categorical_features)


def transform_labels_one(
    config: EnsembleConfig, y_train: np.ndarray | torch.Tensor
) -> np.ndarray:
    """Transform the labels for one ensemble config.
        for both regression or classification.

    Args:
        config: Ensemble config.
        y_train: The unprocessed labels.

    Return: The processed labels.
    """
    if isinstance(config, RegressorEnsembleConfig):
        if config.target_transform is not None:
            # TODO(eddiebergman): Verify this transformer is fitted back in the main
            # process context, otherwise we need some way to return it, possibly
            # by just returning the config
            y_train = config.target_transform.fit_transform(
                y_train.reshape(-1, 1),
            ).ravel()
    elif isinstance(config, ClassifierEnsembleConfig):
        if config.class_permutation is not None:
            y_train = config.class_permutation[y_train]
    else:
        raise ValueError(f"Invalid ensemble config type: {type(config)}")
    return y_train


def fit_preprocessing(
    configs: Sequence[EnsembleConfig],
    X_train: np.ndarray,
    y_train: np.ndarray,
    *,
    random_state: int | np.random.Generator | None,
    cat_ix: list[int],
    n_preprocessing_jobs: int,
    parallel_mode: Literal["block", "as-ready", "in-order"],
) -> Iterator[
    tuple[
        EnsembleConfig,
        SequentialFeatureTransformer,
        np.ndarray,
        np.ndarray,
        list[int],
    ]
]:
    """Fit preprocessing pipelines in parallel.

    Args:
        configs: List of ensemble configurations.
        X_train: Training data.
        y_train: Training target.
        random_state: Random number generator.
        cat_ix: Indices of categorical features.
        n_preprocessing_jobs: Number of worker processes to use.
            If `1`, then the preprocessing is performed in the current process. This
                avoids multiprocessing overheads, but may not be able to full saturate
                the CPU. Note that the preprocessing itself will parallelise over
                multiple cores, so one job is often enough.
            If `>1`, then different estimators are dispatched to different proceses,
                which allows more parallelism but incurs some overhead.
            If `-1`, then creates as many workers as CPU cores. As each worker itself
                uses multiple cores, this is likely too many.
            It is best to select this value by benchmarking.
        parallel_mode:
            Parallel mode to use.

            * `"block"`: Blocks until all workers are done. Returns in order.
            * `"as-ready"`: Returns results as they are ready. Any order.
            * `"in-order"`: Returns results in order, blocking only in the order that
                needs to be returned in.

    Returns:
        Iterator of tuples containing the ensemble configuration, the fitted
        preprocessing pipeline, the transformed training data, the transformed target,
        and the indices of categorical features.
    """
    _, rng = infer_random_state(random_state)

    # Below we set batch_size to auto, but this could be further tuned.
    if SUPPORTS_RETURN_AS:
        return_as = PARALLEL_MODE_TO_RETURN_AS[parallel_mode]
        executor = joblib.Parallel(
            n_jobs=n_preprocessing_jobs,
            return_as=return_as,
            batch_size="auto",
        )
    else:
        executor = joblib.Parallel(n_jobs=n_preprocessing_jobs, batch_size="auto")
    func = partial(fit_preprocessing_one, cat_ix=cat_ix)
    worker_func = joblib.delayed(func)

    seeds = rng.integers(0, np.iinfo(np.int32).max, len(configs))
    yield from executor(  # type: ignore
        [
            worker_func(config, X_train, y_train, seed)
            for config, seed in zip(configs, seeds)
        ],
    )


class DatasetCollectionWithPreprocessing(Dataset):
    """Manages a collection of dataset configurations for lazy processing.

    This class acts as a meta-dataset where each item corresponds to a
    single, complete dataset configuration (e.g., raw features, raw labels,
    preprocessing details defined in `RegressorDatasetConfig` or
    `ClassifierDatasetConfig`). When an item is accessed via `__getitem__`,
    it performs the following steps on the fly:

    1.  Retrieves the specified dataset configuration.
    2.  Splits the raw data into training and testing sets using the provided
        `split_fn` and a random seed derived from `rng`. For regression,
        both raw and pre-standardized targets might be split.
    3.  Fits preprocessors (defined in the dataset configuration's `config`
        attribute) on the *training* data using the `fit_preprocessing`
        utility. This may result in multiple preprocessed versions
        if the configuration specifies an ensemble of preprocessing pipelines.
        For regression we also standardise the target variable.
    4.  Applies the fitted preprocessors to the *testing* features (`x_test_raw`).
    5.  Converts relevant outputs to `torch.Tensor` objects.
    6.  Returns the preprocessed data splits along with other relevant
        information (like raw test data, configs) as a tuple.

    This approach is memory-efficient, especially when dealing with many
    datasets or configurations, as it avoids loading and preprocessing
    everything simultaneously.

    Args:
        split_fn (Callable): A function compatible with scikit-learn's
            `train_test_split` signature (e.g.,
            `sklearn.model_selection.train_test_split`). It's used to split
            the raw data (X, y) into train and test sets. It will receive
            `X`, `y`, and `random_state` as arguments.
        rng: A NumPy random number generator instance
            used for generating the split seed and potentially within the
            preprocessing steps defined in the configs.
        dataset_config_collection: A sequence containing dataset configuration objects.
            Each object must hold the raw data (`X_raw`, `y_raw`), categorical feature
            indices (`cat_ix`), and the specific preprocessing configurations
            (`config`) for that dataset. Regression configs require additional
            fields (`znorm_space_bardist_`).
        n_preprocessing_jobs: The number of workers to use for potentially parallelized
            preprocessing steps (passed to `fit_preprocessing`).

    Attributes:
        configs (Sequence[Union[RegressorDatasetConfig, ClassifierDatasetConfig]]):
            Stores the input dataset configuration collection.
        split_fn (Callable): Stores the splitting function.
        rng (np.random.Generator): Stores the random number generator.
        n_preprocessing_jobs (int): The number of worker processes that will be used for
            the preprocessing.
    """

    def __init__(
        self,
        split_fn: Callable,
        rng: np.random.Generator,
        dataset_config_collection: Sequence[
            RegressorDatasetConfig | ClassifierDatasetConfig
        ],
        n_preprocessing_jobs: int = 1,
    ) -> None:
        self.configs = dataset_config_collection
        self.split_fn = split_fn
        self.rng = rng
        self.n_preprocessing_jobs = n_preprocessing_jobs

    def __len__(self):
        return len(self.configs)

    def __getitem__(self, index: int):  # noqa: C901, PLR0912
        """Retrieves, splits, and preprocesses the dataset config at the index.

        Performs train/test splitting and applies potentially multiple
        preprocessing pipelines defined in the dataset's configuration.

        Args:
            index (int): The index of the dataset configuration in the
                `dataset_config_collection` to process.

        Returns:
            Tuple: A tuple containing the processed data and metadata. Each
                element in the tuple is a list whose length equals the number
                of estimators in the TabPFN ensemble. As such each element
                in the list corresponds to the preprocessed data/configs for a
                single ensemble member.

                The structure depends on the task type derived from the dataset
                configuration object (`RegressorDatasetConfig` or
                `ClassifierDatasetConfig`):

                For **Classification** tasks (`ClassifierDatasetConfig`):
                * `X_trains_preprocessed` (List[torch.Tensor]): List of preprocessed
                  training feature tensors (one per preprocessing pipeline).
                * `X_tests_preprocessed` (List[torch.Tensor]): List of preprocessed
                  test feature tensors (one per preprocessing pipeline).
                * `y_trains_preprocessed` (List[torch.Tensor]): List of preprocessed
                  training target tensors (one per preprocessing pipeline).
                * `y_test_raw` (torch.Tensor): Original, unprocessed test target
                  tensor.
                * `cat_ixs` (List[Optional[List[int]]]): List of categorical feature
                  indices corresponding to each preprocessed X_train/X_test.
                * `conf` (List): The list of preprocessing configurations used for
                  this dataset (usually reflects ensemble settings).

                For **Regression** tasks (`RegressorDatasetConfig`):
                * `X_trains_preprocessed` (List[torch.Tensor]): List of preprocessed
                  training feature tensors.
                * `X_tests_preprocessed` (List[torch.Tensor]): List of preprocessed
                  test feature tensors.
                * `y_trains_preprocessed` (List[torch.Tensor]): List of preprocessed
                  *standardized* training target tensors.
                * `y_test_standardized` (torch.Tensor): *Standardized* test target
                  tensor (derived from `y_full_standardised`).
                * `cat_ixs` (List[Optional[List[int]]]): List of categorical feature
                  indices corresponding to each preprocessed X_train/X_test.
                * `conf` (List): The list of preprocessing configurations used.
                * `raw_space_bardist_` (FullSupportBarDistribution): Binning class
                  for target variable (specific to the regression config). The
                  calculations will be on raw data in raw space.
                * `znorm_space_bardist_` (FullSupportBarDistribution): Binning class for
                  target variable (specific to the regression config). The calculations
                  will be on standardized data in znorm space.
                * `x_test_raw` (torch.Tensor): Original, unprocessed test feature
                  tensor.
                * `y_test_raw` (torch.Tensor): Original, unprocessed test target
                  tensor.

        Raises:
            IndexError: If the index is out of the bounds of the dataset collection.
            ValueError: If the dataset configuration type at the index is not
                        recognized (neither `RegressorDatasetConfig` nor
                        `ClassifierDatasetConfig`).
            AssertionError: If sanity checks during processing fail (e.g.,
                            standardized mean not close to zero in regression).
        """
        if index < 0 or index >= len(self):
            raise IndexError("Index out of bounds.")

        config = self.configs[index]

        # Check type of Dataset Config
        if isinstance(config, RegressorDatasetConfig):
            conf = config.config
            x_full_raw = config.X_raw
            y_full_raw = config.y_raw
            cat_ix = config.cat_ix
            znorm_space_bardist_ = config.znorm_space_bardist_
        elif isinstance(config, ClassifierDatasetConfig):
            conf = config.config
            x_full_raw = config.X_raw
            y_full_raw = config.y_raw
            cat_ix = config.cat_ix
        else:
            raise ValueError(f"Invalid dataset config type: {type(config)}")

        regression_task = isinstance(config, RegressorDatasetConfig)

        x_train_raw, x_test_raw, y_train_raw, y_test_raw = self.split_fn(
            x_full_raw, y_full_raw
        )

        # Compute target variable Z-transform standardization
        # based on statistics of training set
        # Note: Since we compute raw_space_bardist_ here,
        # it is not set as an attribute of the Regressor class
        # This however makes also sense when considering that
        # this attribute changes on every dataset
        if regression_task:
            train_mean = np.mean(y_train_raw)
            train_std = np.std(y_train_raw)
            y_test_standardized = (y_test_raw - train_mean) / train_std
            y_train_standardized = (y_train_raw - train_mean) / train_std
            raw_space_bardist_ = FullSupportBarDistribution(
                znorm_space_bardist_.borders * train_std
                + train_mean  # Inverse normalization back to raw space
            ).float()

        y_train = y_train_standardized if regression_task else y_train_raw

        itr = fit_preprocessing(
            configs=conf,
            X_train=x_train_raw,
            y_train=y_train,
            random_state=self.rng,
            cat_ix=cat_ix,
            n_preprocessing_jobs=self.n_preprocessing_jobs,
            parallel_mode="block",
        )
        (
            configs,
            preprocessors,
            X_trains_preprocessed,
            y_trains_preprocessed,
            cat_ixs,
        ) = list(zip(*itr))
        X_trains_preprocessed = list(X_trains_preprocessed)
        y_trains_preprocessed = list(y_trains_preprocessed)

        ## Process test data for all ensemble estimators.
        X_tests_preprocessed = []
        for _, estim_preprocessor in zip(configs, preprocessors):
            X_tests_preprocessed.append(estim_preprocessor.transform(x_test_raw).X)

        ## Convert to tensors.
        for i in range(len(X_trains_preprocessed)):
            if not isinstance(X_trains_preprocessed[i], torch.Tensor):
                X_trains_preprocessed[i] = torch.as_tensor(
                    X_trains_preprocessed[i], dtype=torch.float32
                )
            if not isinstance(X_tests_preprocessed[i], torch.Tensor):
                X_tests_preprocessed[i] = torch.as_tensor(
                    X_tests_preprocessed[i], dtype=torch.float32
                )
            if not isinstance(y_trains_preprocessed[i], torch.Tensor):
                y_trains_preprocessed[i] = torch.as_tensor(
                    y_trains_preprocessed[i], dtype=torch.float32
                )

        if regression_task and not isinstance(y_test_standardized, torch.Tensor):
            y_test_standardized = torch.from_numpy(y_test_standardized)
            if torch.is_floating_point(y_test_standardized):
                y_test_standardized = y_test_standardized.float()
            else:
                y_test_standardized = y_test_standardized.long()

        x_train_raw = torch.from_numpy(x_train_raw)
        x_test_raw = torch.from_numpy(x_test_raw)
        y_train_raw = torch.from_numpy(y_train_raw)
        y_test_raw = torch.from_numpy(y_test_raw)

        # Also return raw_target variable because of flexiblity
        # in optimisation space -> see examples/
        # Also return corresponding target variable binning
        # classes raw_space_bardist_ and znorm_space_bardist_
        if regression_task:
            return (
                X_trains_preprocessed,
                X_tests_preprocessed,
                y_trains_preprocessed,
                y_test_standardized,
                cat_ixs,
                conf,
                raw_space_bardist_,
                znorm_space_bardist_,
                x_test_raw,
                y_test_raw,
            )

        return (
            X_trains_preprocessed,
            X_tests_preprocessed,
            y_trains_preprocessed,
            y_test_raw,
            cat_ixs,
            conf,
        )



"""This file is taken from PyTorch and modified to work with TabPFN, also
inspired from sklearn's show_versions function. This collects useful debug
information that can be used to report issues.

Usage: import tabpfn; tabpfn.show_versions().
"""

from __future__ import annotations

# mypy: allow-untyped-defs
import datetime
import json
import locale
import os
import platform
import re
import subprocess
import sys
from collections import namedtuple
from pathlib import Path
from tabpfn.settings import settings

try:
    import torch

    TORCH_AVAILABLE = True
except (ImportError, NameError, AttributeError, OSError):
    TORCH_AVAILABLE = False

# System Environment Information
SystemEnv = namedtuple(
    "SystemEnv",
    [
        "torch_version",
        "cuda_compiled_version",
        "gcc_version",
        "clang_version",
        "cmake_version",
        "os",
        "libc_version",
        "python_version",
        "python_platform",
        "is_cuda_available",
        "cuda_runtime_version",
        "cuda_module_loading",
        "nvidia_driver_version",
        "nvidia_gpu_models",
        "cudnn_version",
        "pip_version",  # 'pip' or 'pip3'
        "pip_packages",
        "conda_packages",
        "hip_compiled_version",
        "hip_runtime_version",
        "miopen_runtime_version",
        "caching_allocator_config",
        "is_xnnpack_available",
        "cpu_info",
    ],
)

COMMON_PATTERNS = [
    "torch",
    "numpy",
    "triton",
    "optree",
]

NVIDIA_PATTERNS = [
    "cuda-cudart",
    "cuda-cupti",
    "cuda-libraries",
    "cuda-opencl",
    "cuda-nvrtc",
    "cuda-runtime",
    "cublas",
    "cudnn",
    "cufft",
    "curand",
    "cusolver",
    "cusparse",
    "nccl",
    "nvjitlink",
    "nvtx",
]

CONDA_PATTERNS = [
    "cudatoolkit",
    "soumith",
    "mkl",
    "magma",
]

PIP_PATTERNS = [
    "mypy",
    "flake8",
    "onnx",
]


def _run(command):
    """Return (return-code, stdout, stderr)."""
    shell = isinstance(command, str)
    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=shell,
    )  # nosec: Trusted input;
    # ensure command is not derived from user data.
    raw_output, raw_err = p.communicate()
    rc = p.returncode
    enc = "oem" if _get_platform() == "win32" else locale.getpreferredencoding()
    output = raw_output.decode(enc)
    err = raw_err.decode(enc)
    return rc, output.strip(), err.strip()


def _run_and_read_all(run_lambda, command):
    """Run command using run_lambda; reads and returns entire output if rc is 0."""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    return out


def _run_and_parse_first_match(run_lambda, command, regex):
    """Run command using run_lambda, returns the first regex match if it exists."""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    match = re.search(regex, out)
    if match is None:
        return None
    return match.group(1)


def _run_and_return_first_line(run_lambda, command):
    """Run command using run_lambda and returns first line if output is not empty."""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    return out.split("\n")[0]


def _get_conda_packages(run_lambda, patterns=None):
    if patterns is None:
        patterns = CONDA_PATTERNS + COMMON_PATTERNS + NVIDIA_PATTERNS
    conda = os.environ.get("CONDA_EXE", "conda")
    out = _run_and_read_all(run_lambda, f"{conda} list")
    if out is None:
        return out

    return "\n".join(
        line
        for line in out.splitlines()
        if not line.startswith("#") and any(name in line for name in patterns)
    )


def _get_gcc_version(run_lambda):
    return _run_and_parse_first_match(run_lambda, "gcc --version", r"gcc (.*)")


def _get_clang_version(run_lambda):
    return _run_and_parse_first_match(
        run_lambda,
        "clang --version",
        r"clang version (.*)",
    )


def _get_cmake_version(run_lambda):
    return _run_and_parse_first_match(run_lambda, "cmake --version", r"cmake (.*)")


def _get_nvidia_driver_version(run_lambda):
    if _get_platform() == "darwin":
        cmd = "kextstat | grep -i cuda"
        return _run_and_parse_first_match(
            run_lambda,
            cmd,
            r"com[.]nvidia[.]CUDA [(](.*?)[)]",
        )
    smi = _get_nvidia_smi()
    return _run_and_parse_first_match(run_lambda, smi, r"Driver Version: (.*?) ")


def _get_gpu_info(run_lambda):
    if _get_platform() == "darwin" or (
        TORCH_AVAILABLE
        and hasattr(torch.version, "hip")
        and torch.version.hip is not None
    ):
        if TORCH_AVAILABLE and torch.cuda.is_available():
            if torch.version.hip is not None:
                prop = torch.cuda.get_device_properties(0)
                if hasattr(prop, "gcnArchName"):
                    gcnArch = f" ({prop.gcnArchName})"
                else:
                    gcnArch = "NoGCNArchNameOnOldPyTorch"
            else:
                gcnArch = ""
            return torch.cuda.get_device_name(None) + gcnArch
        return None
    smi = _get_nvidia_smi()
    uuid_regex = re.compile(r" \(UUID: .+?\)")
    rc, out, _ = run_lambda(f"{smi} -L")
    if rc != 0:
        return None
    # Anonymize GPUs by removing their UUID
    return re.sub(uuid_regex, "", out)


def _get_running_cuda_version(run_lambda):
    return _run_and_parse_first_match(run_lambda, "nvcc --version", r"release .+ V(.*)")


def _get_cudnn_version(run_lambda):
    """Return a list of libcudnn.so; it's hard to tell which one is being used."""
    if _get_platform() == "win32":
        system_root = os.environ.get("SYSTEMROOT", "C:\\Windows")
        cuda_path = os.environ.get("CUDA_PATH", "%CUDA_PATH%")
        where_cmd = str(Path(system_root) / "System32" / "where")
        cudnn_cmd = f'{where_cmd} /R "{cuda_path}\\bin" cudnn*.dll'
    elif _get_platform() == "darwin":
        # CUDA libraries and drivers can be found in /usr/local/cuda/.
        cudnn_cmd = "ls /usr/local/cuda/lib/libcudnn*"
    else:
        cudnn_cmd = 'ldconfig -p | grep libcudnn | rev | cut -d" " -f1 | rev'
    rc, out, _ = run_lambda(cudnn_cmd)
    # find will return 1 if there are permission errors or if not found
    if len(out) == 0 or (rc not in (1, 0)):
        cudnn_library = os.environ.get("CUDNN_LIBRARY")
        if cudnn_library is not None and Path(cudnn_library).is_file():
            return os.path.realpath(cudnn_library)
        return None
    files_set = set()
    for file_line in out.split("\n"):
        real_path = os.path.realpath(file_line)  # eliminate symbolic links
        if Path(real_path).is_file():
            files_set.add(real_path)
    if not files_set:
        return None
    files = sorted(files_set)
    if len(files) == 1:
        return files[0]
    result = "\n".join(files)
    return f"Probably one of the following:\n{result}"


def _get_nvidia_smi():
    # Note: nvidia-smi is currently available only on Windows and Linux
    smi = "nvidia-smi"
    if _get_platform() == "win32":
        system_root = os.environ.get("SYSTEMROOT", "C:\\Windows")
        program_files_root = os.environ.get("PROGRAMFILES", "C:\\Program Files")
        legacy_path = str(
            Path(program_files_root) / "NVIDIA Corporation" / "NVSMI" / smi,
        )
        new_path = str(Path(system_root) / "System32" / smi)
        smis = [new_path, legacy_path]
        for candidate_smi in smis:
            if Path(candidate_smi).exists():
                smi = f'"{candidate_smi}"'
                break
    return smi


def _get_cpu_info(run_lambda):
    rc, out, err = 0, "", ""
    if _get_platform() == "linux":
        rc, out, err = run_lambda("lscpu")
    elif _get_platform() == "win32":
        powershell_cmd = (
            'powershell.exe "gwmi -Class Win32_Processor '
            "| Select-Object -Property Name,Manufacturer,Family,Architecture,"
            "ProcessorType,DeviceID,CurrentClockSpeed,MaxClockSpeed,"
            'L2CacheSize,L2CacheSpeed,Revision | ConvertTo-Json"'
        )
        rc, out, err = run_lambda(powershell_cmd)
        if rc == 0:
            lst = []
            try:
                obj = json.loads(out)
                if isinstance(obj, list):
                    for o in obj:
                        lst.append("----------------------")
                        lst.extend(f"{k}: {v}" for k, v in o.items())
                else:
                    lst.extend(f"{k}: {v}" for k, v in obj.items())
            except ValueError as e:
                lst.append(out)
                lst.append(str(e))
            out = "\n".join(lst)
    elif _get_platform() == "darwin":
        rc, out, err = run_lambda("sysctl -n machdep.cpu.brand_string")
    return out if rc == 0 else err


def _get_platform():
    if sys.platform.startswith("linux"):
        return "linux"
    if sys.platform.startswith("win32"):
        return "win32"
    if sys.platform.startswith("cygwin"):
        return "cygwin"
    if sys.platform.startswith("darwin"):
        return "darwin"
    return sys.platform


def _get_mac_version(run_lambda):
    return _run_and_parse_first_match(run_lambda, "sw_vers -productVersion", r"(.*)")


def _get_windows_version(run_lambda):
    powershell_cmd = (
        'powershell.exe "gwmi -Class Win32_OperatingSystem '
        "| Select-Object -Property Caption,OSArchitecture,Version "
        '| ConvertTo-Json"'
    )
    ret = _run_and_read_all(run_lambda, powershell_cmd)
    try:
        obj = json.loads(ret)
        ret = f"{obj['Caption']} ({obj['Version']} {obj['OSArchitecture']})"
    except ValueError as e:
        ret += f"\n{e}"
    return ret


def _get_lsb_version(run_lambda):
    return _run_and_parse_first_match(
        run_lambda,
        "lsb_release -a",
        r"Description:\t(.*)",
    )


def _check_release_file(run_lambda):
    return _run_and_parse_first_match(
        run_lambda,
        "cat /etc/*-release",
        r'PRETTY_NAME="(.*)"',
    )


def _get_os(run_lambda):
    from platform import machine

    plat = _get_platform()
    os_description = None

    if plat in ("win32", "cygwin"):
        os_description = _get_windows_version(run_lambda)
    elif plat == "darwin":
        version = _get_mac_version(run_lambda)
        os_description = f"macOS {version} ({machine()})" if version else None
    elif plat == "linux":
        desc = _get_lsb_version(run_lambda)
        if desc is None:
            desc = _check_release_file(run_lambda)
        os_description = f"{desc} ({machine()})" if desc else f"{plat} ({machine()})"
    else:
        os_description = plat

    return os_description


def _get_python_platform():
    return platform.platform()


def _get_libc_version():
    if _get_platform() != "linux":
        return "N/A"
    return "-".join(platform.libc_ver())


def _get_pip_packages(run_lambda, patterns=None):
    """Return `pip list` output.

    Note: Will also find conda-installed pytorch and numpy packages.
    """
    if patterns is None:
        patterns = PIP_PATTERNS + COMMON_PATTERNS + NVIDIA_PATTERNS

    pip_version = "pip3" if sys.version_info[0] >= 3 else "pip"

    os.environ["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
    out = _run_and_read_all(
        run_lambda,
        [sys.executable, "-mpip", "list", "--format=freeze"],
    )
    if out is None:
        # Happens when pip is not available or when the command fails
        return pip_version, ""
    filtered_out = "\n".join(
        line for line in out.splitlines() if any(name in line for name in patterns)
    )

    return pip_version, filtered_out


def _get_cachingallocator_config():
    return settings.pytorch.pytorch_cuda_alloc_conf


def _get_cuda_module_loading_config():
    if TORCH_AVAILABLE and torch.cuda.is_available():
        torch.cuda.init()
        return os.environ.get("CUDA_MODULE_LOADING", "")
    return "N/A"


def _is_xnnpack_available():
    if TORCH_AVAILABLE:
        import torch.backends.xnnpack

        return str(torch.backends.xnnpack.enabled)  # type: ignore[attr-defined]
    return "N/A"


def _get_env_info():
    """Collect environment information to aid in debugging.

    The returned environment information contains details on torch version,
    whether it is a debug build or not, the CUDA compiled version, gcc version,
    clang version, and cmake version, as well as the operating system and libc
    version. It also includes the python version, python platform, CUDA
    availability, CUDA runtime version, and CUDA module loading configuration.
    Further, it details the GPU model and configuration, Nvidia driver version,
    cuDNN version, pip version, and versions of relevant pip and conda packages.
    Additionally, HIP runtime version, MIOpen runtime version, caching allocator
    config, XNNPACK availability, and CPU information are provided.

    Returns:
        SystemEnv (namedtuple): A tuple containing various environment details
        and system information.
    """
    run_lambda = _run
    pip_version, pip_list_output = _get_pip_packages(run_lambda)
    if TORCH_AVAILABLE:
        version_str = torch.__version__
        cuda_available_str = str(torch.cuda.is_available())
        cuda_version_str = torch.version.cuda
        if (
            not hasattr(torch.version, "hip") or torch.version.hip is None
        ):  # cuda version
            hip_compiled_version = hip_runtime_version = miopen_runtime_version = "N/A"
        else:  # HIP version

            def get_version_or_na(cfg, prefix):
                _lst = [s.rsplit(None, 1)[-1] for s in cfg if prefix in s]
                return _lst[0] if _lst else "N/A"

            cfg = torch._C._show_config().split("\n")
            hip_runtime_version = get_version_or_na(cfg, "HIP Runtime")
            miopen_runtime_version = get_version_or_na(cfg, "MIOpen")
            cuda_version_str = "N/A"
            hip_compiled_version = torch.version.hip
    else:
        version_str = cuda_available_str = cuda_version_str = "N/A"
        hip_compiled_version = hip_runtime_version = miopen_runtime_version = "N/A"

    sys_version = sys.version.replace("\n", " ")
    conda_packages = _get_conda_packages(run_lambda)
    return SystemEnv(
        torch_version=version_str,
        python_version=f"{sys_version} ({sys.maxsize.bit_length() + 1}-bit runtime)",
        python_platform=_get_python_platform(),
        is_cuda_available=cuda_available_str,
        cuda_compiled_version=cuda_version_str,
        cuda_runtime_version=_get_running_cuda_version(run_lambda),
        cuda_module_loading=_get_cuda_module_loading_config(),
        nvidia_gpu_models=_get_gpu_info(run_lambda),
        nvidia_driver_version=_get_nvidia_driver_version(run_lambda),
        cudnn_version=_get_cudnn_version(run_lambda),
        hip_compiled_version=hip_compiled_version,
        hip_runtime_version=hip_runtime_version,
        miopen_runtime_version=miopen_runtime_version,
        pip_version=pip_version,
        pip_packages=pip_list_output,
        conda_packages=conda_packages,
        os=_get_os(run_lambda),
        libc_version=_get_libc_version(),
        gcc_version=_get_gcc_version(run_lambda),
        clang_version=_get_clang_version(run_lambda),
        cmake_version=_get_cmake_version(run_lambda),
        caching_allocator_config=_get_cachingallocator_config(),
        is_xnnpack_available=_is_xnnpack_available(),
        cpu_info=_get_cpu_info(run_lambda),
    )


env_info_fmt = """
PyTorch version: {torch_version}
CUDA used to build PyTorch: {cuda_compiled_version}
ROCM used to build PyTorch: {hip_compiled_version}

OS: {os}
GCC version: {gcc_version}
Clang version: {clang_version}
CMake version: {cmake_version}
Libc version: {libc_version}

Python version: {python_version}
Python platform: {python_platform}
Is CUDA available: {is_cuda_available}
CUDA runtime version: {cuda_runtime_version}
CUDA_MODULE_LOADING set to: {cuda_module_loading}
GPU models and configuration: {nvidia_gpu_models}
Nvidia driver version: {nvidia_driver_version}
cuDNN version: {cudnn_version}
HIP runtime version: {hip_runtime_version}
MIOpen runtime version: {miopen_runtime_version}
Is XNNPACK available: {is_xnnpack_available}

CPU:
{cpu_info}
""".strip()


def _replace_nones(dct, replacement="Could not collect"):
    for key in dct:
        if dct[key] is None:
            dct[key] = replacement
    return dct


def _replace_bools(dct, true="Yes", false="No"):
    for key in dct:
        if dct[key] is True:
            dct[key] = true
        elif dct[key] is False:
            dct[key] = false
    return dct


def _prepend(text, tag="[prepend]"):
    lines = text.split("\n")
    updated_lines = [f"{tag}{line}" for line in lines]
    return "\n".join(updated_lines)


def _replace_if_empty(text, replacement="No relevant packages"):
    if text is not None and len(text) == 0:
        return replacement
    return text


def _maybe_start_on_next_line(string):
    # If `string` is multiline, prepend a \n to it.
    if string is not None and len(string.split("\n")) > 1:
        return f"\n{string}\n"
    return string


def _pretty_str(envinfo):
    mutable_dict = envinfo._asdict()

    # If nvidia_gpu_models is multiline, start on the next line
    mutable_dict["nvidia_gpu_models"] = _maybe_start_on_next_line(
        envinfo.nvidia_gpu_models,
    )

    # If the machine doesn't have CUDA, report some fields as 'No CUDA'
    dynamic_cuda_fields = [
        "cuda_runtime_version",
        "nvidia_gpu_models",
        "nvidia_driver_version",
    ]
    all_cuda_fields = [*dynamic_cuda_fields, "cudnn_version"]
    all_dynamic_cuda_fields_missing = all(
        mutable_dict[field] is None for field in dynamic_cuda_fields
    )
    if (
        TORCH_AVAILABLE
        and not torch.cuda.is_available()
        and all_dynamic_cuda_fields_missing
    ):
        for field in all_cuda_fields:
            mutable_dict[field] = "No CUDA"
        if envinfo.cuda_compiled_version is None:
            mutable_dict["cuda_compiled_version"] = "None"

    # Replace True with Yes, False with No
    mutable_dict = _replace_bools(mutable_dict)

    # Replace all None objects with 'Could not collect'
    mutable_dict = _replace_nones(mutable_dict)

    # If either of these are '', replace with 'No relevant packages'
    mutable_dict["pip_packages"] = _replace_if_empty(mutable_dict["pip_packages"])
    mutable_dict["conda_packages"] = _replace_if_empty(mutable_dict["conda_packages"])

    # Tag conda and pip packages with a prefix
    # If they were previously None, they'll show up as ie '[conda] Could not collect'
    if mutable_dict["pip_packages"]:
        mutable_dict["pip_packages"] = _prepend(
            mutable_dict["pip_packages"],
            f"[{envinfo.pip_version}] ",
        )
    if mutable_dict["conda_packages"]:
        mutable_dict["conda_packages"] = _prepend(
            mutable_dict["conda_packages"],
            "[conda] ",
        )
    mutable_dict["cpu_info"] = envinfo.cpu_info
    return env_info_fmt.format(**mutable_dict)


def _get_deps_info():
    """Overview of the installed version of main dependencies.

    Returns:
    -------
    deps_info: dict
        version information on relevant Python libraries
    """
    deps = [
        "torch",
        "numpy",
        "scipy",
        "pandas",
        "scikit-learn",
        "typing_extensions",
        "einops",
        "huggingface-hub",
    ]

    from importlib.metadata import version as get_version  # Rename to avoid shadowing

    try:
        __version__ = get_version("tabpfn")
    except ImportError:
        __version__ = "unknown"

    deps_info = {
        "tabpfn": __version__,
    }

    from importlib.metadata import PackageNotFoundError

    for modname in deps:
        try:
            deps_info[modname] = get_version(modname)  # Use renamed function
        except PackageNotFoundError:
            deps_info[modname] = "Not Found"  # Use a string instead of None
    return deps_info


def display_debug_info():
    """Print useful debugging information."""
    print("Collecting system and dependency information...")
    sys_info = _get_env_info()._replace(pip_packages="", conda_packages="")
    print(
        _pretty_str(sys_info).replace(
            "Versions of relevant libraries:\n"
            "[pip3] No relevant packages\n"
            "[conda] No relevant packages\n",
            "",
        )
    )

    # Print dependency versions in a clear section
    print("\nDependency Versions:")
    print("-" * 20)
    deps_info = _get_deps_info()
    for pkg, version in deps_info.items():
        print(f"{pkg}: {version or 'Not installed'}")

    # Check for crash dumps
    if (
        TORCH_AVAILABLE
        and hasattr(torch, "utils")
        and hasattr(torch.utils, "_crash_handler")
    ):
        minidump_dir = torch.utils._crash_handler.DEFAULT_MINIDUMP_DIR
        if sys.platform == "linux" and Path(minidump_dir).exists():
            dumps = [
                str(Path(minidump_dir) / dump) for dump in os.listdir(minidump_dir)
            ]
            if dumps:  # Only proceed if there are dumps
                latest = max(dumps, key=lambda p: Path(p).stat().st_ctime)
                ctime = Path(latest).stat().st_ctime
                crash_time = datetime.datetime.fromtimestamp(ctime).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
                print(f"\nLatest crash dump found from: {crash_time}")
                print(f"Location: {latest}")



"""Test to verify that telemetry is disabled during test execution."""

from __future__ import annotations

import os


def test_telemetry_disabled():
    """Verify that TABPFN_DISABLE_TELEMETRY is set to 1 during tests."""
    assert os.environ.get("TABPFN_DISABLE_TELEMETRY") == "1", (
        "TABPFN_DISABLE_TELEMETRY should be set to '1' during test execution. "
        "This ensures telemetry is disabled for all tests."
    )



from __future__ import annotations

import unittest
from functools import partial
from typing import Literal
from unittest.mock import patch

import numpy as np
import pytest
import sklearn
import torch
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from tabpfn import TabPFNClassifier
from tabpfn.preprocessing import (
    ClassifierEnsembleConfig,
    DatasetCollectionWithPreprocessing,
)
from tabpfn.utils import meta_dataset_collator

from .utils import get_pytest_devices

rng = np.random.default_rng(42)

devices = get_pytest_devices()

fit_modes = [
    "batched",
    "fit_preprocessors",
]

inference_precision_methods: list[torch.types._dtype | Literal["autocast", "auto"]] = [
    "auto",
    torch.float64,
]
estimators = [1, 2]

param_order = [
    "n_estimators",
    "device",
    "fit_mode",
    "inference_precision",
]

default_config = {
    "n_estimators": 1,
    "device": "auto",
    "fit_mode": "batched",
    "inference_precision": "auto",
}

param_values: dict[str, list] = {
    "n_estimators": estimators,
    "device": devices,
    "fit_mode": fit_modes,
    "inference_precision": inference_precision_methods,
}

combinations = [tuple(default_config[param] for param in param_order)]
for param in param_order:
    for value in param_values[param]:
        if value != default_config[param]:
            config = default_config.copy()
            config[param] = value
            combinations.append(tuple(config[param] for param in param_order))


@pytest.fixture(scope="module")
def synthetic_data():
    X = rng.normal(size=(100, 4)).astype(np.float32)
    y = rng.integers(0, 3, size=100).astype(np.float32)
    return X, y


# Fixture: synthetic collection of datasets (list of (X, y) tuples)
@pytest.fixture(scope="module")
def uniform_synthetic_dataset_collection():
    datasets = []
    for _ in range(3):
        X = rng.normal(size=(30, 3)).astype(np.float32)
        y = rng.integers(0, 3, size=30)
        datasets.append((X, y))
    return datasets


@pytest.fixture(scope="module")
def classification_data():
    """Generate simple classification data."""
    X, y = make_classification(
        n_samples=20,
        n_features=5,
        n_informative=3,
        n_redundant=0,
        n_classes=3,
        random_state=42,
    )
    y = y - y.min()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=42, stratify=y
    )
    return X_train, X_test, y_train, y_test


@pytest.fixture(params=devices)
def classifier_instance(request) -> TabPFNClassifier:
    """Provides a basic classifier instance, parameterized by device."""
    device = request.param
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")
    # Uses defaults similar to the original fixture, but with device param
    return TabPFNClassifier(
        n_estimators=2,
        device=device,
        random_state=42,
        inference_precision=torch.float32,
        fit_mode="batched",
        differentiable_input=False,
    )


# --- Helper to Create Classifier (for fully parameterized tests) ---
def create_classifier(
    n_estimators: int,
    device: str,
    fit_mode: str,
    inference_precision: torch.types._dtype | Literal["autocast", "auto"],
    **kwargs,
) -> TabPFNClassifier:
    """Instantiates classifier with common parameters."""
    if device == "cpu" and inference_precision == "autocast":
        pytest.skip("Unsupported combination: CPU with 'autocast'")
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    default_kwargs = {"random_state": 42}
    default_kwargs.update(kwargs)

    return TabPFNClassifier(
        n_estimators=n_estimators,
        device=device,
        fit_mode=fit_mode,
        inference_precision=inference_precision,
        **default_kwargs,
    )


# Fixture: synthetic collection of datasets (list of (X, y) tuples)
# where dataset size and number of classes vary.
@pytest.fixture
def variable_synthetic_dataset_collection():
    """Fixture: synthetic collection of datasets
    (list of (X, y) tuples)
    where dataset size and number of classes vary.
    """
    datasets = []
    dataset_sizes = [10, 20, 30]
    class_counts = [2, 4, 6]
    n_features = 3
    for size, n_classes in zip(dataset_sizes, class_counts):
        X = rng.normal(size=(size, n_features)).astype(np.float32)
        y = rng.integers(0, n_classes, size=size)
        datasets.append((X, y))
    return datasets


@pytest.mark.parametrize(param_order, combinations)
def test_tabpfn_classifier_finetuning_loop(
    n_estimators,
    device,
    fit_mode,
    inference_precision,
    synthetic_data,
) -> None:
    X, y = synthetic_data
    X_train, _X_test, y_train, _y_test = train_test_split(
        X, y, test_size=0.3, random_state=42
    )

    clf = create_classifier(
        n_estimators,
        device,
        fit_mode,
        inference_precision,
        random_state=2,
        differentiable_input=False,
    )

    datasets_list = clf.get_preprocessed_datasets(
        X_train, y_train, train_test_split, 100
    )
    lossfn = torch.nn.NLLLoss()
    batch_size = 1
    my_dl_train = DataLoader(
        datasets_list, batch_size=batch_size, collate_fn=meta_dataset_collator
    )

    if inference_precision == torch.float64:
        # Expect to raise a ValueError in
        # TODO: check that it fails with the right error
        pass

    elif fit_mode in [
        "fit_preprocessors",
        "fit_with_cache",
        "low_memory",
    ]:
        # TODO: check that it fails with the right error
        pass

    else:
        for data_batch in my_dl_train:
            X_tr, X_te, y_tr, y_te_raw, cat_ixs, confs = data_batch

            # --- Fit and Predict ---
            clf.fit_from_preprocessed(X_tr, y_tr, cat_ixs, confs)
            preds = clf.forward(X_te)

            # --- Basic Shape Checks ---
            assert preds.ndim == 3, f"Expected 3D output, got {preds.shape}"
            assert preds.shape[0] == X_te[0].shape[0]
            assert preds.shape[0] == y_te_raw.shape[0]
            assert preds.shape[1] == clf.n_classes_, "Class count mismatch"
            assert len(X_te) == clf.n_estimators
            assert len(X_tr) == clf.n_estimators
            assert len(y_tr) == clf.n_estimators

            # --- Loss Calculation and Backward Pass ---
            log_preds = torch.log(preds + 1e-12)
            # Target shape needs adjustment for NLLLoss (B, N) -> (B*N)
            # Prediction shape (B, Cls, N) -> (B*N, Cls)
            target = y_te_raw.to(preds.device).long()
            target = target.view(-1)  # Flatten target to (B*N)
            # Permute and flatten preds: (B, Cls, N) -> (B, N, Cls) -> (B*N, Cls)
            log_preds_permuted = log_preds.permute(0, 2, 1).contiguous()
            log_preds_flat = log_preds_permuted.view(-1, clf.n_classes_)

            assert log_preds_flat.ndim == 2
            assert target.ndim == 1
            assert log_preds_flat.shape[0] == target.shape[0]  # Total samples match

            loss = lossfn(log_preds_flat, target)
            assert torch.isfinite(loss).all(), f"Loss is not finite: {loss.item()}"

            # --- Gradient Check ---
            assert hasattr(clf, "models_")
            assert clf.models_ is not None
            clf.models_[0].zero_grad()
            loss.backward()

            gradients_found = False
            for param in clf.models_[0].parameters():
                if (
                    param.requires_grad
                    and param.grad is not None
                    and param.grad.abs().sum().item() > 1e-12
                ):
                    gradients_found = True
                    break
            assert gradients_found, "No non-zero gradients found."
            break  # Only test one batch


def test_get_preprocessed_datasets_basic():
    X = rng.normal(size=(100, 4)).astype(np.float32)
    y = rng.integers(0, 3, size=100)

    clf = TabPFNClassifier()
    # This should return a DatasetCollectionWithPreprocessing
    dataset = clf.get_preprocessed_datasets(X, y, split_fn=train_test_split)
    assert hasattr(dataset, "__getitem__")
    assert hasattr(dataset, "__len__")
    assert len(dataset) > 0
    item = dataset[0]
    assert isinstance(item, tuple)
    assert len(item) == 6


def test_datasetcollectionwithpreprocessing_classification_single_dataset(
    synthetic_data, classifier_instance: TabPFNClassifier
) -> None:
    X_raw, y_raw = synthetic_data
    clf = classifier_instance
    n_estimators = clf.n_estimators
    test_size = 0.3

    split_fn = partial(train_test_split, test_size=test_size, shuffle=True)
    dataset_collection = clf.get_preprocessed_datasets(
        X_raw, y_raw, split_fn=split_fn, max_data_size=None
    )

    assert isinstance(dataset_collection, DatasetCollectionWithPreprocessing)
    assert len(dataset_collection) == 1, "Collection should contain one dataset config"

    item_index = 0
    processed_dataset_item = dataset_collection[item_index]

    assert isinstance(processed_dataset_item, tuple)
    assert len(processed_dataset_item) == 6, (
        "Item tuple should have 4 elements for classification"
    )

    (
        X_trains_preprocessed,
        _X_tests_preprocessed,
        _y_trains_preprocessed,
        y_test_raw_tensor,
        _cat_ixs,
        _returned_ensemble_configs,
    ) = processed_dataset_item

    assert isinstance(X_trains_preprocessed, list)
    assert len(X_trains_preprocessed) == n_estimators
    n_samples_total = X_raw.shape[0]
    expected_n_test = int(np.floor(n_samples_total * test_size))
    expected_n_train = n_samples_total - expected_n_test
    assert y_test_raw_tensor.shape == (expected_n_test,)
    assert X_trains_preprocessed[0].shape[0] == expected_n_train


def test_datasetcollectionwithpreprocessing_classification_multiple_datasets(
    uniform_synthetic_dataset_collection, classifier_instance: TabPFNClassifier
) -> None:
    """Test DatasetCollectionWithPreprocessing
    using a collection of multiple synthetic datasets.
    """
    datasets = uniform_synthetic_dataset_collection
    clf = classifier_instance
    n_estimators = clf.n_estimators
    test_size = 0.3
    split_fn = partial(train_test_split, test_size=test_size, shuffle=True)

    X_list = [X for X, _ in datasets]
    y_list = [y for _, y in datasets]

    dataset_collection = clf.get_preprocessed_datasets(
        X_list, y_list, split_fn=split_fn, max_data_size=None
    )

    assert isinstance(dataset_collection, DatasetCollectionWithPreprocessing)
    assert len(dataset_collection) == len(datasets), (
        "Collection should contain one item per dataset"
    )

    for item_index in range(len(datasets)):
        processed_dataset_item = dataset_collection[item_index]
        assert isinstance(processed_dataset_item, tuple)
        assert len(processed_dataset_item) == 6, (
            "Item tuple should have 6 elements for classification"
        )
        (
            X_trains_preprocessed,
            _X_tests_preprocessed,
            _y_trains_preprocessed,
            y_test_raw_tensor,
            _cat_ixs,
            _returned_ensemble_configs,
        ) = processed_dataset_item
        assert isinstance(X_trains_preprocessed, list)
        assert len(X_trains_preprocessed) == n_estimators
        n_samples_total = X_list[item_index].shape[0]
        expected_n_test = int(np.floor(n_samples_total * test_size))
        expected_n_train = n_samples_total - expected_n_test
        assert y_test_raw_tensor.shape == (expected_n_test,)
        assert X_trains_preprocessed[0].shape[0] == expected_n_train


def test_dataset_and_collator_with_dataloader_uniform(
    uniform_synthetic_dataset_collection, classifier_instance
) -> None:
    X_list = [X for X, _ in uniform_synthetic_dataset_collection]
    y_list = [y for _, y in uniform_synthetic_dataset_collection]
    dataset_collection = classifier_instance.get_preprocessed_datasets(
        X_list, y_list, train_test_split, 100
    )
    batch_size = 1
    dl = DataLoader(
        dataset_collection,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        # Should be tuple with X_trains, X_tests, y_trains, y_tests, cat_ixs, confs
        assert isinstance(batch, tuple)
        X_trains, _X_tests, y_trains, _y_tests, _cat_ixs, _confs = batch
        for est_tensor in X_trains:
            assert isinstance(est_tensor, torch.Tensor), (
                "Each estimator's batch should be a tensor."
            )
            assert est_tensor.shape[0] == batch_size
        for est_tensor in y_trains:
            assert isinstance(est_tensor, torch.Tensor), (
                "Each estimator's batch should be a tensor for labels."
            )
            assert est_tensor.shape[0] == batch_size
        break  # Only check one batch


def test_classifier_dataset_and_collator_batches_type(
    variable_synthetic_dataset_collection, classifier_instance
):
    """Test that the batches returned by the dataset and collator
    are of the correct type.
    """
    X_list = [X for X, _ in variable_synthetic_dataset_collection]
    y_list = [y for _, y in variable_synthetic_dataset_collection]
    dataset_collection = classifier_instance.get_preprocessed_datasets(
        X_list, y_list, train_test_split, 100
    )
    batch_size = 1
    dl = DataLoader(
        dataset_collection,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        assert isinstance(batch, tuple)
        X_trains, _X_tests, y_trains, _y_tests, cat_ixs, confs = batch
        for est_tensor in X_trains:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        for est_tensor in y_trains:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        assert isinstance(cat_ixs, list)
        for conf in confs:
            for c in conf:
                assert isinstance(c, ClassifierEnsembleConfig)
        break


def test_get_preprocessed_datasets_multiple_datasets(classifier_instance):
    # Provide lists of datasets
    X1 = rng.standard_normal((10, 4))
    y1 = rng.integers(0, 2, size=10)
    X2 = rng.standard_normal((8, 4))
    y2 = rng.integers(0, 2, size=8)
    datasets = classifier_instance.get_preprocessed_datasets(
        [X1, X2], [y1, y2], split_fn=None
    )
    assert hasattr(datasets, "__getitem__")
    assert len(datasets) == 2


def test_get_preprocessed_datasets_categorical_features(classifier_instance):
    # One categorical column (e.g., int-coded)
    X = np.array([[0, 1.2, 3.4], [1, 2.3, 4.5], [0, 0.1, 2.2], [2, 1.1, 3.3]])
    y = np.array([0, 1, 0, 1])
    # Specify categorical_features_indices
    classifier_instance.categorical_features_indices = [0]
    datasets = classifier_instance.get_preprocessed_datasets(X, y, split_fn=None)
    # Should not raise, and should process categorical features
    assert hasattr(datasets, "__getitem__")
    # Optionally, check that categorical indices are stored or used


def test_forward_runs(classifier_instance, classification_data):
    """Ensure predict_proba_tensor runs OK after standard fit."""
    X_train, X_test, y_train, _y_test = classification_data
    clf = classifier_instance
    clf.fit_mode = "low_memory"
    clf.fit(X_train, y_train)
    preds = clf.forward(
        torch.tensor(X_test, dtype=torch.float32), use_inference_mode=True
    )
    # Check output shape and probability sum
    assert preds.ndim == 2, f"Expected 2D output, got {preds.shape}"
    assert preds.shape[0] == X_test.shape[0], "Mismatch in test sample count"
    assert preds.shape[1] == clf.n_classes_, "Mismatch in class count"
    probs_sum = preds.sum(dim=1)
    assert torch.allclose(probs_sum, torch.ones_like(probs_sum), atol=1e-5), (
        "Probabilities do not sum to 1"
    )


def test_fit_from_preprocessed_runs(classifier_instance, classification_data) -> None:
    """Verify fit_from_preprocessed runs
    using prepared data and produces
    valid predictions.
    """
    X_train, _X_test, y_train, _y_test = classification_data
    clf = classifier_instance

    split_fn = partial(train_test_split, test_size=0.3, random_state=42)

    datasets_list = clf.get_preprocessed_datasets(X_train, y_train, split_fn, 100)
    batch_size = 1
    dl = DataLoader(
        datasets_list, batch_size=batch_size, collate_fn=meta_dataset_collator
    )

    for data_batch in dl:
        X_trains, X_tests, y_trains, y_tests, cat_ixs, confs = data_batch
        clf.fit_from_preprocessed(X_trains, y_trains, cat_ixs, confs)
        preds = clf.forward(X_tests)
        assert preds.ndim == 3, f"Expected 3D output, got {preds.shape}"
        assert preds.shape[0] == X_tests[0].shape[0]
        assert preds.shape[0] == y_tests.shape[0]
        assert preds.shape[1] == clf.n_classes_

        # TODO: verify number of classes, "Mismatch in class count"
        probs_sum = preds.sum(dim=1)
        assert torch.allclose(probs_sum, torch.ones_like(probs_sum), atol=1e-5), (
            "Probabilities do not sum to 1"
        )
        break  # Only need to check one batch for this test


class TestTabPFNClassifierPreprocessingInspection(unittest.TestCase):
    def test_finetuning_consistency_preprocessing_classifier(self):
        """Tests the consistency between standard preprocessing (fit -> predict_proba)
        and the fine-tuning preprocessing pipeline
        (get_preprocessed_datasets -> fit_from_preprocessed
            -> forward)
        for the TabPFNClassifier by comparing the tensors entering the internal model.
        """
        # --- Test Parameters ---
        test_set_size = 0.3
        common_seed = 42
        n_total = 50  # Increased slightly for more robust testing
        n_features = 8
        n_classes = 2  # Use a specific number of classes
        n_informative = 5  # For make_classification
        n_estimators = 1  # Keep N=1 for easier direct comparison of tensors

        # --- 1. Setup ---
        X, y = sklearn.datasets.make_classification(
            n_samples=n_total,
            n_features=n_features,
            n_informative=n_informative,
            n_redundant=n_features - n_informative,
            n_classes=n_classes,
            n_clusters_per_class=1,  # Simpler structure
            random_state=common_seed,
        )
        splitfn = partial(
            train_test_split,
            test_size=test_set_size,
            random_state=common_seed,
            shuffle=False,  # Keep False for consistent splitting
        )
        X_train_raw, X_test_raw, y_train_raw, _y_test_raw = splitfn(X, y)

        # Initialize two classifiers with the necessary modes
        clf_standard = TabPFNClassifier(
            n_estimators=n_estimators,
            device="auto",
            random_state=common_seed,
            fit_mode="fit_preprocessors",  # A standard mode that preprocesses on fit
        )
        # 'batched' mode is required for get_preprocessed_datasets
        #  and fit_from_preprocessed
        clf_batched = TabPFNClassifier(
            n_estimators=n_estimators,
            device="auto",
            random_state=common_seed,
            fit_mode="batched",
        )

        # --- 2. Path 1: Standard fit -> predict_proba -> Capture Tensor ---

        clf_standard.fit(X_train_raw, y_train_raw)
        # Ensure the internal model attribute exists after fit
        assert all(
            [
                hasattr(clf_standard, "models_"),
                hasattr(clf_standard.models_[0], "forward"),
            ]
        ), "Standard classifier models_ or models_[0].forward not found after fit."

        tensor_p1_full = None
        # Patch the standard classifier's *internal model's* forward method
        # The internal model typically receives the combined train+test sequence
        with patch.object(
            clf_standard.models_[0], "forward", wraps=clf_standard.models_[0].forward
        ) as mock_forward_p1:
            _ = clf_standard.predict_proba(X_test_raw)
            assert mock_forward_p1.called, "Standard models_[0].forward was not called."

            # Capture the tensor input 'x' (usually the second positional argument)
            call_args_list = mock_forward_p1.call_args_list
            assert len(call_args_list) > 0, (
                "No calls recorded for standard models_[0].forward."
            )
            if len(call_args_list[0].args) > 1:
                tensor_p1_full = call_args_list[0].args[0]
                tensor_p1_full = mock_forward_p1.call_args.args[0]

            else:
                self.fail(
                    f"Standard models_[0].forward call had "
                    f"unexpected arguments: {call_args_list[0].args}"
                )

        assert tensor_p1_full is not None, (
            "Failed to capture tensor from standard path."
        )
        # Shape might be [1, N_Total, Features+1] or similar. Check the actual shape.
        # Example assertion: Check if the sequence length matches n_total
        assert tensor_p1_full.shape[0] == n_total, (
            f"Path 1 tensor sequence length ({tensor_p1_full.shape[0]})"
            f"does not match n_total ({n_total}). Shape was {tensor_p1_full.shape}"
        )

        # FT Workflow (get_prep -> fit_prep -> predict_prep -> Capture Tensor) ---
        # Step 3a: Get preprocessed datasets using the *full* dataset
        # Requires fit_mode='batched' on clf_batched
        # Make sure default max_data_size is large enough.
        datasets_list = clf_batched.get_preprocessed_datasets(
            X,
            y,
            splitfn,  # Use the full X, y and the split function
        )
        assert len(datasets_list) > 0, "get_preprocessed_datasets returned empty list."

        dataloader = DataLoader(
            datasets_list,
            batch_size=1,
            collate_fn=meta_dataset_collator,
            shuffle=False,
        )
        try:
            data_batch = next(iter(dataloader))
        except StopIteration:
            self.fail("DataLoader yielded no batches.")

        try:
            (X_trains_p2, X_tests_p2, y_trains_p2, _, cat_ixs_p2, confs_p2, *_) = (
                data_batch
            )
        except ValueError as e:
            self.fail(
                f"Failed to unpack data batch from DataLoader."
                f"Structure might be different. Error: {e}. Batch content: {data_batch}"
            )

        clf_batched.fit_from_preprocessed(
            X_trains_p2, y_trains_p2, cat_ixs_p2, confs_p2
        )
        assert all(
            [
                hasattr(clf_batched, "models_"),
                hasattr(clf_batched.models_[0], "forward"),
            ]
        ), (
            "Batched classifier models_ or models_[0].forward not"
            "found after fit_from_preprocessed."
        )

        # Step 3c: Call forward and capture the input tensor
        # to the *internal transformer model*
        tensor_p2_full = None
        # Patch the *batched* classifier's internal model's forward method
        with patch.object(
            clf_batched.models_[0], "forward", wraps=clf_batched.models_[0].forward
        ) as mock_forward_p2:
            _ = clf_batched.forward(X_tests_p2)
            assert mock_forward_p2.called, "Batched models_[0].forward was not called."

            # Capture the tensor input 'x' (assuming same argument position as Path 1)
            call_args_list = mock_forward_p2.call_args_list
            assert len(call_args_list) > 0, (
                "No calls recorded for batched models_[0].forward."
            )
            if len(call_args_list[0].args) > 1:
                tensor_p2_full = mock_forward_p2.call_args.args[0]
            else:
                self.fail(
                    f"Batched models_[0].forward call had "
                    f"unexpected arguments: {call_args_list[0].args}"
                )

        assert tensor_p2_full is not None, "Failed to capture tensor from batched path."
        # The internal model in this path should
        # also receive the full sequence if n_estimators=1
        # and the dataloader yielded the full split.
        assert tensor_p2_full.shape[0] == n_total, (
            f"Path 2 tensor sequence length ({tensor_p2_full.shape[0]}) "
            f"does not match n_total ({n_total}). Shape was {tensor_p2_full.shape}"
        )

        # --- 4. Comparison (Path 1 vs Path 2) ---

        # Ensure tensors are on the same device (CPU) for comparison
        tensor_p1_full = tensor_p1_full.cpu()
        tensor_p2_full = tensor_p2_full.cpu()

        # Squeeze dimensions of size 1
        # E.g., if shape is [1, N_Total, Features+1], squeeze the first dim
        if tensor_p1_full.shape[0] == 1:
            p1_squeezed = tensor_p1_full.squeeze(0)
        else:
            p1_squeezed = tensor_p1_full

        if tensor_p2_full.shape[0] == 1:
            p2_squeezed = tensor_p2_full.squeeze(0)
        else:
            p2_squeezed = tensor_p2_full

        # Final check of shapes after potential squeeze
        assert p1_squeezed.shape == p2_squeezed.shape, (
            "Shapes of final model input tensors mismatch after squeeze. "
        )

        # Visual inspection snippet

        # Perform numerical comparison using torch.allclose
        # Use a reasonably small tolerance. Preprocessing should be near-identical.
        # Floating point ops might introduce tiny differences.
        atol = 1e-6
        rtol = 1e-5
        tensors_match = torch.allclose(p1_squeezed, p2_squeezed, atol=atol, rtol=rtol)

        if not tensors_match:
            diff = torch.abs(p1_squeezed - p2_squeezed)
            # Find where they differ most
            _max_diff_val, max_diff_idx = torch.max(diff.flatten(), dim=0)
            np.unravel_index(max_diff_idx.item(), p1_squeezed.shape)

        # Assertion: The final tensors fed to the model sh
        assert tensors_match, "Mismatch between final model input tensors."



from __future__ import annotations

import unittest
from functools import partial
from typing import Literal
from unittest.mock import patch

import numpy as np
import pytest
import sklearn
import torch
from sklearn.model_selection import train_test_split
from torch.optim import Adam
from torch.utils.data import DataLoader

from tabpfn import TabPFNRegressor
from tabpfn.architectures.base.bar_distribution import (
    BarDistribution,
    FullSupportBarDistribution,
)
from tabpfn.preprocessing import RegressorEnsembleConfig
from tabpfn.utils import meta_dataset_collator

from .utils import get_pytest_devices

rng = np.random.default_rng(42)

devices = get_pytest_devices()

fit_modes = [
    "batched",
    "fit_preprocessors",
]
inference_precision_methods: list[torch.types._dtype | Literal["autocast", "auto"]] = [
    "auto",
    torch.float64,
]
estimators = [1, 2]
optimization_spaces_values = ["raw_label_space", "preprocessed"]

param_order = [
    "n_estimators",
    "device",
    "fit_mode",
    "inference_precision",
    "optimization_space",
]

default_config = {
    "n_estimators": 1,
    "device": "cpu",
    "fit_mode": "batched",
    "inference_precision": "auto",
    "optimization_space": "raw_label_space",
}

param_values: dict[str, list] = {
    "n_estimators": estimators,
    "device": devices,
    "fit_mode": fit_modes,
    "inference_precision": inference_precision_methods,
    "optimization_space": optimization_spaces_values,
}

combinations = [tuple(default_config[p] for p in param_order)]
for param_name in param_order:
    for value in param_values[param_name]:
        if value != default_config[param_name]:
            current_config = default_config.copy()
            current_config[param_name] = value
            combinations.append(tuple(current_config[p] for p in param_order))


@pytest.fixture(scope="module")
def synthetic_regression_data():
    """Generate synthetic regression data."""
    X = rng.normal(size=(30, 4)).astype(np.float32)
    # Generate continuous target variable
    y = (X @ rng.normal(size=4)).astype(np.float32)
    # Add to previous as line too long otherwise
    y += rng.normal(size=30).astype(np.float32) * 0.1
    return X, y


@pytest.fixture(params=devices)
def ft_regressor_instance(request) -> TabPFNRegressor:
    """Provides a basic regressor instance, parameterized by device."""
    device = request.param
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")
    return TabPFNRegressor(
        n_estimators=2,
        device=device,
        random_state=42,
        inference_precision=torch.float32,
        fit_mode="batched",
        differentiable_input=False,
    )


@pytest.fixture(params=devices)
def std_regressor_instance(request) -> TabPFNRegressor:
    """Provides a basic regressor instance, parameterized by device."""
    device = request.param
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")
    return TabPFNRegressor(
        n_estimators=2,
        device=device,
        random_state=42,
        inference_precision=torch.float32,
        fit_mode="low_memory",
        differentiable_input=False,
    )


def create_regressor(
    n_estimators: int,
    device: str,
    fit_mode: str,
    inference_precision: torch.types._dtype | Literal["autocast", "auto"],
    **kwargs,
) -> TabPFNRegressor:
    """Instantiates regressor with common parameters."""
    if device == "cpu" and inference_precision == "autocast":
        pytest.skip("Unsupported combination: CPU with 'autocast'")
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    default_kwargs = {"random_state": 42}
    default_kwargs.update(kwargs)

    return TabPFNRegressor(
        n_estimators=n_estimators,
        device=device,
        fit_mode=fit_mode,
        inference_precision=inference_precision,
        **default_kwargs,
    )


# --- Tests ---


def test_regressor_dataset_and_collator_batches_type(
    synthetic_regression_data, ft_regressor_instance
):
    """Test that the batches returned by the dataset and collator
    are of the correct type.
    """
    X, y = synthetic_regression_data
    dataset_collection = ft_regressor_instance.get_preprocessed_datasets(
        X, y, train_test_split, 100
    )
    batch_size = 1
    dl = DataLoader(
        dataset_collection,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        assert isinstance(batch, tuple)
        (
            X_trains_preprocessed,
            _X_tests_preprocessed,
            y_trains_preprocessed,
            _y_test_standardized,
            cat_ixs,
            confs,
            raw_space_bardist_,
            bar_distribution,
            _x_test_raw,
            _y_test_raw,
        ) = batch
        for est_tensor in X_trains_preprocessed:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        for est_tensor in y_trains_preprocessed:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        assert isinstance(cat_ixs, list)
        for conf in confs:
            for c in conf:
                assert isinstance(c, RegressorEnsembleConfig)
        for ren_crit in raw_space_bardist_:
            assert isinstance(ren_crit, FullSupportBarDistribution)
        for bar_dist in bar_distribution:
            assert isinstance(bar_dist, BarDistribution)
        break


@pytest.mark.parametrize(param_order, combinations)
def test_tabpfn_regressor_finetuning_loop(
    n_estimators,
    device,
    fit_mode,
    inference_precision,
    optimization_space,
    synthetic_regression_data,
) -> None:
    X, y = synthetic_regression_data
    X_train, _X_test, y_train, _y_test = train_test_split(
        X, y, test_size=0.3, random_state=42
    )

    reg = create_regressor(
        n_estimators,
        device,
        fit_mode,
        inference_precision,
        random_state=2,
        differentiable_input=False,
    )

    datasets_list = reg.get_preprocessed_datasets(
        X_train, y_train, train_test_split, 100
    )

    batch_size = 1
    my_dl_train = DataLoader(
        datasets_list, batch_size=batch_size, collate_fn=meta_dataset_collator
    )

    optim_impl = Adam(reg.models_[0].parameters(), lr=1e-5)

    if inference_precision == torch.float64:
        pass
        # TODO: check that it fails with the right error

    elif fit_mode in [
        "fit_preprocessors",
        "fit_with_cache",
        "low_memory",
    ]:
        # TODO: check that it fails with the right error
        pass
    else:
        for data_batch in my_dl_train:
            optim_impl.zero_grad()

            (
                X_trains_preprocessed,
                X_tests_preprocessed,
                y_trains_preprocessed,
                y_test_standardized,
                cat_ixs,
                confs,
                raw_space_bardist_,
                _bar_distribution,
                _batch_x_test_raw,
                batch_y_test_raw,
            ) = data_batch

            reg.fit_from_preprocessed(
                X_trains_preprocessed, y_trains_preprocessed, cat_ixs, confs
            )

            reg.raw_space_bardist_ = raw_space_bardist_[0]

            averaged_pred_logits, _, _ = reg.forward(X_tests_preprocessed)

            # --- Basic Shape Checks ---
            assert averaged_pred_logits.ndim == 3, (
                f"Expected 3D output, got {averaged_pred_logits.shape}"
            )

            # Batch Size
            assert averaged_pred_logits.shape[0] == batch_y_test_raw.shape[0]
            assert averaged_pred_logits.shape[0] == batch_size
            assert averaged_pred_logits.shape[0] == X_tests_preprocessed[0].shape[0]
            assert averaged_pred_logits.shape[0] == y_test_standardized.shape[0]

            # N_samples
            assert averaged_pred_logits.shape[1] == batch_y_test_raw.shape[1]
            assert averaged_pred_logits.shape[1] == y_test_standardized.shape[1]

            # N_bins
            n_borders_bardist = reg.znorm_space_bardist_.borders.shape[0]
            assert averaged_pred_logits.shape[2] == n_borders_bardist - 1
            n_borders_norm_crit = reg.raw_space_bardist_.borders.shape[0]
            assert averaged_pred_logits.shape[2] == n_borders_norm_crit - 1

            assert len(X_tests_preprocessed) == reg.n_estimators
            assert len(X_trains_preprocessed) == reg.n_estimators
            assert len(y_trains_preprocessed) == reg.n_estimators
            assert reg.models_ is not None, "Model not initialized after fit"
            assert hasattr(reg, "znorm_space_bardist_"), (
                "Regressor missing 'znorm_space_bardist_' attribute after fit"
            )
            assert hasattr(reg, "raw_space_bardist_"), (
                "Regressor missing 'raw_space_bardist_' attribute after fit"
            )
            assert reg.znorm_space_bardist_ is not None, (
                "reg.znorm_space_bardist_ is None"
            )

            lossfn = None
            if optimization_space == "raw_label_space":
                lossfn = reg.raw_space_bardist_
            elif optimization_space == "preprocessed":
                lossfn = reg.znorm_space_bardist_
            else:
                raise ValueError("Need to define optimization space")

            nll_loss_per_sample = lossfn(
                averaged_pred_logits, batch_y_test_raw.to(device)
            )
            loss = nll_loss_per_sample.mean()

            # --- Gradient Check ---
            loss.backward()
            optim_impl.step()

            assert torch.isfinite(loss).all(), f"Loss is not finite: {loss.item()}"

            gradients_found = False
            for param in reg.models_[0].parameters():
                if (
                    param.requires_grad
                    and param.grad is not None
                    and param.grad.abs().sum().item() > 1e-12
                ):
                    gradients_found = True
                    break
            assert gradients_found, "No non-zero gradients found."

            reg.models_[0].zero_grad()
            break  # Only test one batch


def test_finetuning_consistency_bar_distribution(
    std_regressor_instance, ft_regressor_instance, synthetic_regression_data
):
    """Tests if predict() output matches the output derived from
    get_preprocessed_datasets -> fit_from_preprocessed -> forward() -> post-processing,
    when no actual fine-tuning occurs.
    """
    common_seed = 10
    test_set_size = 0.2

    reg_standard = std_regressor_instance
    reg_batched = ft_regressor_instance

    if reg_standard.device != reg_batched.device:
        pytest.skip("Devices do not match.")

    x_full_raw, y_full_raw = synthetic_regression_data

    splitfn = partial(
        train_test_split,
        test_size=test_set_size,
        random_state=common_seed,
        shuffle=False,
    )

    X_train_raw, X_test_raw, y_train_raw, y_test_raw = splitfn(x_full_raw, y_full_raw)

    reg_standard.fit(X_train_raw, y_train_raw)
    reg_standard.predict(X_test_raw, output_type="mean")

    datasets_list = reg_batched.get_preprocessed_datasets(
        x_full_raw, y_full_raw, splitfn, max_data_size=1000
    )

    batch_size = 1
    dataloader = DataLoader(
        datasets_list,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
        shuffle=False,
    )
    data_batch = next(iter(dataloader))
    (
        X_trains_preprocessed,
        _X_tests_preprocessed,
        y_trains_preprocessed,
        y_test_standardized,
        cat_ixs,
        confs,
        raw_space_bardist_,
        _bar_distribution,
        batch_x_test_raw,
        batch_y_test_raw,
    ) = data_batch

    np.testing.assert_allclose(
        batch_y_test_raw.flatten().detach().cpu().numpy(),
        y_test_raw,
        rtol=1e-5,
        atol=1e-5,
    )

    reg_batched.fit_from_preprocessed(
        X_trains_preprocessed, y_trains_preprocessed, cat_ixs, confs
    )

    mean = np.mean(y_train_raw)
    std = np.std(y_train_raw)
    y_train_std_ = std.item() + 1e-20
    y_train_mean_ = mean.item()
    y_standardised_investigated = (y_test_raw - y_train_mean_) / y_train_std_

    np.testing.assert_allclose(
        y_test_standardized[0].flatten().detach().cpu().numpy(),
        y_standardised_investigated,
        rtol=1e-5,
        atol=1e-5,
    )

    np.testing.assert_allclose(
        batch_x_test_raw[0].detach().cpu().numpy(),
        X_test_raw,
        rtol=1e-5,
        atol=1e-5,
    )

    raw_space_bardist_ = raw_space_bardist_[0]
    reg_batched.raw_space_bardist_ = raw_space_bardist_

    torch.testing.assert_close(
        raw_space_bardist_.borders,
        reg_batched.raw_space_bardist_.borders,
        rtol=1e-5,
        atol=1e-5,
        msg="Renormalized criterion borders do not match.",
    )

    torch.testing.assert_close(
        raw_space_bardist_.borders,
        reg_standard.raw_space_bardist_.borders,
        rtol=1e-5,  # Standard float tolerance
        atol=1e-5,
        msg="Renormalized criterion borders do not match.",
    )

    torch.testing.assert_close(
        reg_standard.raw_space_bardist_.borders,
        reg_batched.raw_space_bardist_.borders,
        rtol=1e-5,  # Standard float tolerance
        atol=1e-5,
        msg="Renormalized criterion borders do not match.",
    )

    torch.testing.assert_close(
        reg_standard.znorm_space_bardist_.borders,
        reg_batched.znorm_space_bardist_.borders,
        rtol=1e-5,  # Standard float tolerance
        atol=1e-5,
        msg="Bar distribution borders do not match.",
    )


# ----------------


class TestTabPFNPreprocessingInspection(unittest.TestCase):
    def test_finetuning_consistency_preprocessing_regressor(self):
        """In order to test the consistency of our FineTuning code
        and the preprocessing code, we will test the consistency
        of the preprocessed datasets. We do this by checking
        comparing the tensors that enter the internal transformer
        model.
        """
        test_set_size = 0.3
        common_seed = 42
        n_total = 20
        n_features = 10
        n_estimators = 1

        X, y = sklearn.datasets.make_regression(
            n_samples=n_total, n_features=n_features, random_state=common_seed
        )
        splitfn = partial(
            train_test_split,
            test_size=test_set_size,
            random_state=common_seed,
            shuffle=False,  # Keep False for consistent results if slicing were needed
        )
        X_train_raw, X_test_raw, y_train_raw, _ = splitfn(X, y)

        # Initialize two regressors with the inference and FineTuning
        reg_standard = TabPFNRegressor(
            n_estimators=n_estimators,
            device="auto",
            random_state=common_seed,
            fit_mode="fit_preprocessors",  # Example standard mode
        )
        reg_batched = TabPFNRegressor(
            n_estimators=n_estimators,
            device="auto",
            random_state=common_seed,
            fit_mode="batched",  # Mode compatible with get_preprocessed_datasets
        )

        # --- 2. Path 1: Standard fit -> predict -> Capture Tensor ---
        reg_standard.fit(X_train_raw, y_train_raw)
        assert hasattr(reg_standard, "models_")
        assert hasattr(reg_standard.models_[0], "forward")

        tensor_p1_full = None
        # Patch the standard regressor's internal model's forward method
        with patch.object(
            reg_standard.models_[0], "forward", wraps=reg_standard.models_[0].forward
        ) as mock_forward_p1:
            _ = reg_standard.predict(X_test_raw)  # Trigger the patched method
            assert mock_forward_p1.called
            # Capture the tensor input to the internal model
            tensor_p1_full = mock_forward_p1.call_args.args[0]

        assert tensor_p1_full is not None
        # Standard path's internal model receives the combined train+test sequence
        assert tensor_p1_full.shape[0] == n_total

        # --- 3. Path 3: FT Full Workflow ---
        # (get_prep -> fit_prep -> forward -> Capture Tensor)

        datasets_list = reg_batched.get_preprocessed_datasets(
            X, y, splitfn, max_data_size=1000
        )

        # Fit FT regressor
        dataloader = DataLoader(
            datasets_list,
            batch_size=1,
            collate_fn=meta_dataset_collator,
            shuffle=False,
        )
        data_batch = next(iter(dataloader))
        (
            X_trains_p2,
            X_tests_p2,
            y_trains_p2,
            _,
            cat_ixs_p2,
            confs_p2,
            _,
            _,
            _,
            _,
        ) = data_batch
        reg_batched.fit_from_preprocessed(
            X_trains_p2, y_trains_p2, cat_ixs_p2, confs_p2
        )
        assert hasattr(reg_batched, "models_")
        assert hasattr(reg_batched.models_[0], "forward")

        # Step 3c: Call forward and capture the input tensor to the *internal model*
        tensor_p3_full = None
        # Patch the *batched* regressor's internal model's forward method
        with patch.object(
            reg_batched.models_[0], "forward", wraps=reg_batched.models_[0].forward
        ) as mock_forward_p3:
            # Pass the list of preprocessed test tensors obtained earlier
            _ = reg_batched.forward(X_tests_p2)
            assert mock_forward_p3.called
            # Capture the tensor input to the internal model
            tensor_p3_full = mock_forward_p3.call_args.args[0]

        assert tensor_p3_full is not None
        # As confirmed before, the internal model in this path
        # also receives the full sequence
        assert tensor_p3_full.shape[0] == n_total

        # --- 4. Comparison (Path 1 vs Path 3) ---

        # Compare the two full tensors captured from the input to models_[0].forward
        # Squeeze dimensions of size 1 for direct comparison
        # shapes should be [N_Total, Features+1]
        p1_squeezed = tensor_p1_full.squeeze()
        p3_squeezed = tensor_p3_full.squeeze()

        assert p1_squeezed.shape == p3_squeezed.shape, (
            "Shapes of final model input tensors mismatch."
        )

        atol = 1e-6
        tensors_match = torch.allclose(p1_squeezed, p3_squeezed, atol=atol)

        assert tensors_match, "Mismatch between preprocessed model input tensors."



"""TabPFNRegressor class.

!!! example
    ```python
    import sklearn.datasets
    from tabpfn import TabPFNRegressor

    model = TabPFNRegressor()
    X, y = sklearn.datasets.make_regression(n_samples=50, n_features=10)

    model.fit(X, y)
    predictions = model.predict(X)
    ```
"""

#  Copyright (c) Prior Labs GmbH 2025.

from __future__ import annotations

import logging
import typing
import warnings
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Literal, Union
from typing_extensions import Self, TypedDict, deprecated, overload

import numpy as np
import torch
from sklearn import config_context
from sklearn.base import (
    BaseEstimator,
    RegressorMixin,
    TransformerMixin,
    check_is_fitted,
)
from tabpfn_common_utils.telemetry import track_model_call
from tabpfn_common_utils.telemetry.interactive import ping

from tabpfn.architectures.base.bar_distribution import FullSupportBarDistribution
from tabpfn.base import (
    RegressorModelSpecs,
    check_cpu_warning,
    create_inference_engine,
    determine_precision,
    get_preprocessed_datasets_helper,
    initialize_model_variables_helper,
)
from tabpfn.constants import REGRESSION_CONSTANT_TARGET_BORDER_EPSILON, ModelVersion
from tabpfn.inference import InferenceEngine, InferenceEngineBatchedNoPreprocessing
from tabpfn.model_loading import (
    ModelSource,
    get_cache_dir,
    load_fitted_tabpfn_model,
    save_fitted_tabpfn_model,
)
from tabpfn.preprocessing import (
    DatasetCollectionWithPreprocessing,
    EnsembleConfig,
    RegressorEnsembleConfig,
)
from tabpfn.preprocessors import get_all_reshape_feature_distribution_preprocessors
from tabpfn.preprocessors.preprocessing_helpers import get_ordinal_encoder
from tabpfn.utils import (
    DevicesSpecification,
    fix_dtypes,
    get_embeddings,
    infer_categorical_features,
    infer_random_state,
    process_text_na_dataframe,
    transform_borders_one,
    translate_probs_across_borders,
    validate_X_predict,
    validate_Xy_fit,
)

if TYPE_CHECKING:
    import numpy.typing as npt
    from sklearn.compose import ColumnTransformer
    from sklearn.pipeline import Pipeline
    from torch.types import _dtype

    from tabpfn.architectures.base.memory import MemorySavingMode
    from tabpfn.architectures.interface import Architecture, ArchitectureConfig
    from tabpfn.constants import XType, YType
    from tabpfn.inference import InferenceEngine
    from tabpfn.inference_config import InferenceConfig

    try:
        from sklearn.base import Tags
    except ImportError:
        Tags = Any


# --- Prediction Output Types and Constants ---

# 1. Tuples for runtime validation and internal logic.
# These are defined directly as tuples of strings for immediate clarity.
_OUTPUT_TYPES_BASIC = ("mean", "median", "mode")
_OUTPUT_TYPES_QUANTILES = ("quantiles",)
_OUTPUT_TYPES = _OUTPUT_TYPES_BASIC + _OUTPUT_TYPES_QUANTILES
_OUTPUT_TYPES_COMPOSITE = ("full", "main")
_USABLE_OUTPUT_TYPES = _OUTPUT_TYPES + _OUTPUT_TYPES_COMPOSITE


# 2. Type aliases for static type checking and IDE support.
OutputType = Literal["mean", "median", "mode", "quantiles", "full", "main"]
"""The type hint for the `output_type` parameter in `predict`."""


class MainOutputDict(TypedDict):
    """Specifies the return structure for `output_type="main"`."""

    mean: np.ndarray
    median: np.ndarray
    mode: np.ndarray
    quantiles: list[np.ndarray]


class FullOutputDict(MainOutputDict):
    """Specifies the return structure for `output_type="full"`."""

    criterion: FullSupportBarDistribution
    logits: torch.Tensor


RegressionResultType = Union[
    np.ndarray, list[np.ndarray], MainOutputDict, FullOutputDict
]
"""The type hint for the return value of the `predict` method."""


class TabPFNRegressor(RegressorMixin, BaseEstimator):
    """TabPFNRegressor class."""

    configs_: list[ArchitectureConfig]
    """The configurations of the loaded models to be used for inference.

    The concrete type of these configs is defined by the architectures in use and should
    be inspected at runtime, but they will be subclasses of ArchitectureConfig.
    """

    models_: list[Architecture]
    """The loaded models to be used for inference.

    The models can be different PyTorch modules, but will be subclasses of Architecture.
    """

    inference_config_: InferenceConfig
    """Additional configuration of inference for expert users."""

    devices_: tuple[torch.device, ...]
    """The devices determined to be used.

    The devices are determined based on the `device` argument to the constructor, and
    the devices available on the system. If multiple devices are listed, currently only
    the first is used for inference.
    """

    feature_names_in_: npt.NDArray[Any]
    """The feature names of the input data.

    May not be set if the input data does not have feature names,
    such as with a numpy array.
    """

    n_features_in_: int
    """The number of features in the input data used during `fit()`."""

    inferred_categorical_indices_: list[int]
    """The indices of the columns that were inferred to be categorical,
    as a product of any features deemed categorical by the user and what would
    work best for the model.
    """

    n_outputs_: Literal[1]  # We only support single output
    """The number of outputs the model supports. Only 1 for now"""

    znorm_space_bardist_: FullSupportBarDistribution
    """The bar distribution of the target variable, used by the model.
    This is the bar distribution in the normalized target space.
    """

    raw_space_bardist_: FullSupportBarDistribution
    """The bar distribution in the raw target space, used for computing the
    predictions."""

    use_autocast_: bool
    """Whether torch's autocast should be used."""

    forced_inference_dtype_: _dtype | None
    """The forced inference dtype for the model based on `inference_precision`."""

    executor_: InferenceEngine
    """The inference engine used to make predictions."""

    preprocessor_: ColumnTransformer
    """The column transformer used to preprocess the input data to be numeric."""

    def __init__(  # noqa: PLR0913
        self,
        *,
        n_estimators: int = 8,
        categorical_features_indices: Sequence[int] | None = None,
        softmax_temperature: float = 0.9,
        average_before_softmax: bool = False,
        model_path: str
        | Path
        | list[str]
        | list[Path]
        | Literal["auto"]
        | RegressorModelSpecs
        | list[RegressorModelSpecs] = "auto",
        device: DevicesSpecification = "auto",
        ignore_pretraining_limits: bool = False,
        inference_precision: _dtype | Literal["autocast", "auto"] = "auto",
        fit_mode: Literal[
            "low_memory",
            "fit_preprocessors",
            "fit_with_cache",
            "batched",
        ] = "fit_preprocessors",
        memory_saving_mode: MemorySavingMode = "auto",
        random_state: int | np.random.RandomState | np.random.Generator | None = 0,
        n_jobs: Annotated[int | None, deprecated("Use n_preprocessing_jobs")] = None,
        n_preprocessing_jobs: int = 1,
        inference_config: dict | InferenceConfig | None = None,
        differentiable_input: bool = False,
    ) -> None:
        """Construct a TabPFN regressor.

        This constructs a regressor using the latest model and settings. If you would
        like to use a previous model version, use `create_default_for_version()`
        instead. You can also use `model_path` to specify a particular model.

        Args:
            n_estimators:
                The number of estimators in the TabPFN ensemble. We aggregate the
                predictions of `n_estimators`-many forward passes of TabPFN.
                Each forward pass has (slightly) different input data. Think of this
                as an ensemble of `n_estimators`-many "prompts" of the input data.

            categorical_features_indices:
                The indices of the columns that are suggested to be treated as
                categorical. If `None`, the model will infer the categorical columns.
                If provided, we might ignore some of the suggestion to better fit the
                data seen during pre-training.

                !!! note
                    The indices are 0-based and should represent the data passed to
                    `.fit()`. If the data changes between the initializations of the
                    model and the `.fit()`, consider setting the
                    `.categorical_features_indices` attribute after the model was
                    initialized and before `.fit()`.

            softmax_temperature:
                The temperature for the softmax function. This is used to control the
                confidence of the model's predictions. Lower values make the model's
                predictions more confident. This is only applied when predicting during
                a post-processing step. Set `softmax_temperature=1.0` for no effect.

            average_before_softmax:
                Only used if `n_estimators > 1`. Whether to average the predictions of
                the estimators before applying the softmax function. This can help to
                improve predictive performance when there are many classes or when
                calibrating the model's confidence. This is only applied when
                predicting during a post-processing.

                - If `True`, the predictions are averaged before applying the softmax
                  function. Thus, we average the logits of TabPFN and then apply the
                  softmax.
                - If `False`, the softmax function is applied to each set of logits.
                  Then, we average the resulting probabilities of each forward pass.

            model_path:
                The path to the TabPFN model file, i.e., the pre-trained weights.

                - If `"auto"`, the model will be downloaded upon first use. This
                  defaults to your system cache directory, but can be overwritten
                  with the use of an environment variable `TABPFN_MODEL_CACHE_DIR`.
                - If a path or a string of a path, the model will be loaded from
                  the user-specified location if available, otherwise it will be
                  downloaded to this location.

            device:
                The device(s) to use for inference.

                If "auto": a single device is selected based on availability in the
                following order of priority: "cuda:0", "mps", "cpu".

                To manually select a single device: specify a PyTorch device string e.g.
                "cuda:1". See PyTorch's documentation for information about supported
                devices.

                To use several GPUs: specify a list of PyTorch GPU device strings, e.g.
                ["cuda:0", "cuda:1"]. This can dramatically speed up inference for
                larger datasets, by executing the estimators in parallel on the GPUs.

            ignore_pretraining_limits:
                Whether to ignore the pre-training limits of the model. The TabPFN
                models have been pre-trained on a specific range of input data. If the
                input data is outside of this range, the model may not perform well.
                You may ignore our limits to use the model on data outside the
                pre-training range.

                - If `True`, the model will not raise an error if the input data is
                  outside the pre-training range. Also suppresses error when using
                  the model with more than 1000 samples on CPU.
                - If `False`, you can use the model outside the pre-training range, but
                  the model could perform worse.

                !!! note

                    For version 2.5, the pre-training limits are:

                    - 50_000 samples/rows
                    - 2_000 features/columns (Note that for more than 500 features we
                        subsample 500 features per estimator. It is therefore important
                        to use a sufficiently large number of `n_estimators`.)

            device:
                The device to use for inference with TabPFN. If `"auto"`, the device is
                `"cuda"` if available, otherwise `"cpu"`.

                See PyTorch's documentation on devices for more information about
                supported devices.

            inference_precision:
                The precision to use for inference. This can dramatically affect the
                speed and reproducibility of the inference. Higher precision can lead to
                better reproducibility but at the cost of speed. By default, we optimize
                for speed and use torch's mixed-precision autocast. The options are:

                - If `torch.dtype`, we force precision of the model and data to be
                  the specified torch.dtype during inference. This can is particularly
                  useful for reproducibility. Here, we do not use mixed-precision.
                - If `"autocast"`, enable PyTorch's mixed-precision autocast. Ensure
                  that your device is compatible with mixed-precision.
                - If `"auto"`, we determine whether to use autocast or not depending on
                  the device type.

            fit_mode:
                Determine how the TabPFN model is "fitted". The mode determines how the
                data is preprocessed and cached for inference. This is unique to an
                in-context learning foundation model like TabPFN, as the "fitting" is
                technically the forward pass of the model. The options are:

                - If `"low_memory"`, the data is preprocessed on-demand during inference
                  when calling `.predict()` or `.predict_proba()`. This is the most
                  memory-efficient mode but can be slower for large datasets because
                  the data is (repeatedly) preprocessed on-the-fly.
                  Ideal with low GPU memory and/or a single call to `.fit()` and
                  `.predict()`.
                - If `"fit_preprocessors"`, the data is preprocessed and cached once
                  during the `.fit()` call. During inference, the cached preprocessing
                  (of the training data) is used instead of re-computing it.
                  Ideal with low GPU memory and multiple calls to `.predict()` with
                  the same training data.
                - If `"fit_with_cache"`, the data is preprocessed and cached once during
                  the `.fit()` call like in `fit_preprocessors`. Moreover, the
                  transformer key-value cache is also initialized, allowing for much
                  faster inference on the same data at a large cost of memory.
                  Ideal with very high GPU memory and multiple calls to `.predict()`
                  with the same training data.
                - If `"batched"`, the already pre-processed data is iterated over in
                  batches. This can only be done after the data has been preprocessed
                  with the get_preprocessed_datasets function. This is primarily used
                  only for inference with the InferenceEngineBatchedNoPreprocessing
                  class in Fine-Tuning. The fit_from_preprocessed() function sets this
                  attribute internally.

            memory_saving_mode:
                Enable GPU/CPU memory saving mode. This can both avoid out-of-memory
                errors and improve fit+predict speed by reducing memory pressure.

                It saves memory by automatically batching certain model computations
                within TabPFN.

                - If "auto": memory saving mode is enabled/disabled automatically based
                    on a heuristic
                - If True/False: memory saving mode is forced enabled/disabled.

                If speed is important to your application, you may wish to manually tune
                this option by comparing the time taken for fit+predict with it set to
                False and True.

                !!! warning
                    This does not batch the original input data. We still recommend to
                    batch the test set as necessary if you run out of memory.

            random_state:
                Controls the randomness of the model. Pass an int for reproducible
                results and see the scikit-learn glossary for more information.
                If `None`, the randomness is determined by the system when calling
                `.fit()`.

                !!! warning
                    We depart from the usual scikit-learn behavior in that by default
                    we provide a fixed seed of `0`.

                !!! note
                    Even if a seed is passed, we cannot always guarantee reproducibility
                    due to PyTorch's non-deterministic operations and general numerical
                    instability. To get the most reproducible results across hardware,
                    we recommend using a higher precision as well (at the cost of a
                    much higher inference time). Likewise, for scikit-learn, consider
                    passing `USE_SKLEARN_16_DECIMAL_PRECISION=True` as kwarg.

            n_jobs:
                Deprecated, use `n_preprocessing_jobs` instead.
                This parameter never had any effect.

            n_preprocessing_jobs:
                The number of worker processes to use for the preprocessing.

                If `1`, the preprocessing will be performed in the current process,
                parallelised across multiple CPU cores. If `>1` and `n_estimators > 1`,
                then different estimators will be dispatched to different processes.

                We strongly recommend setting this to 1, which has the lowest overhead
                and can often fully utilise the CPU. Values >1 can help if you have lots
                of CPU cores available, but can also be slower.

            inference_config:
                For advanced users, additional advanced arguments that adjust the
                behavior of the model interface.
                See [tabpfn.inference_config.InferenceConfig][] for details and options.

                - If `None`, the default InferenceConfig is used.
                - If `dict`, the key-value pairs are used to update the default
                  `InferenceConfig`. Raises an error if an unknown key is passed.
                - If `InferenceConfig`, the object is used as the configuration.

            differentiable_input:
                If true, preprocessing attempts to be end-to-end differentiable.
                Less relevant for standard regression fine-tuning compared to
                prompt-tuning.
        """
        super().__init__()
        self.n_estimators = n_estimators
        self.categorical_features_indices = categorical_features_indices
        self.softmax_temperature = softmax_temperature
        self.average_before_softmax = average_before_softmax
        self.model_path = model_path
        self.device = device
        self.ignore_pretraining_limits = ignore_pretraining_limits
        self.inference_precision: torch.dtype | Literal["autocast", "auto"] = (
            inference_precision
        )
        self.fit_mode: Literal["low_memory", "fit_preprocessors", "batched"] = fit_mode
        self.memory_saving_mode = memory_saving_mode
        self.random_state = random_state
        self.inference_config = inference_config
        self.differentiable_input = differentiable_input

        if n_jobs is not None:
            warnings.warn(
                "TabPFNRegressor(n_jobs=...) is deprecated and has no effect. "
                "Use `n_preprocessing_jobs` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        self.n_jobs = n_jobs
        self.n_preprocessing_jobs = n_preprocessing_jobs

        # Ping the usage service if telemetry enabled
        ping()

    @classmethod
    def create_default_for_version(cls, version: ModelVersion, **overrides) -> Self:
        """Construct a regressor that uses the given version of the model.

        In addition to selecting the model, this also configures certain settings to the
        default values associated with this model version.

        Any kwargs will override the default settings.
        """
        if version == ModelVersion.V2:
            options = {
                "model_path": str(
                    get_cache_dir() / ModelSource.get_regressor_v2().default_filename
                ),
                "n_estimators": 8,
                "softmax_temperature": 0.9,
            }
        elif version == ModelVersion.V2_5:
            options = {
                "model_path": str(
                    get_cache_dir() / ModelSource.get_regressor_v2_5().default_filename
                ),
                "n_estimators": 8,
                "softmax_temperature": 0.9,
            }
        else:
            raise ValueError(f"Unknown version: {version}")

        options.update(overrides)

        return cls(**options)

    @property
    def model_(self) -> Architecture:
        """The model used for inference.

        This is set after the model is loaded and initialized.
        """
        if not hasattr(self, "models_"):
            raise ValueError(
                "The model has not been initialized yet. Please initialize the model "
                "before using the `model_` property."
            )
        if len(self.models_) > 1:
            raise ValueError(
                "The `model_` property is not supported when multiple models are used. "
                "Use `models_` instead."
            )
        return self.models_[0]

    @property
    def norm_bardist_(self) -> FullSupportBarDistribution:
        """WARNING: DEPRECATED. Please use `raw_space_bardist_` instead.
        This attribute will be removed in a future version.
        """
        warnings.warn(
            "`norm_bardist_` is deprecated and will be removed in a future version. "
            "Please use `raw_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.raw_space_bardist_

    @norm_bardist_.setter
    def norm_bardist_(self, value: FullSupportBarDistribution) -> None:
        warnings.warn(
            "`norm_bardist_` is deprecated and will be removed in a future version. "
            "Please use `raw_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.raw_space_bardist_ = value

    @property
    def bardist_(self) -> FullSupportBarDistribution:
        """WARNING: DEPRECATED. Please use `znorm_space_bardist_` instead.
        This attribute will be removed in a future version.
        """
        warnings.warn(
            "`bardist_` is deprecated and will be removed in a future version. "
            "Please use `znorm_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.znorm_space_bardist_

    @bardist_.setter
    def bardist_(self, value: FullSupportBarDistribution) -> None:
        warnings.warn(
            "`bardist_` is deprecated and will be removed in a future version. "
            "Please use `znorm_space_bardist_` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.znorm_space_bardist_ = value

    # TODO: We can remove this from scikit-learn lower bound of 1.6
    def _more_tags(self) -> dict[str, Any]:
        return {
            "allow_nan": True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "regressor"
        return tags

    def get_preprocessed_datasets(
        self,
        X_raw: XType | list[XType],
        y_raw: YType | list[YType],
        split_fn: Callable,
        max_data_size: None | int = 10000,
        *,
        equal_split_size: bool = True,
    ) -> DatasetCollectionWithPreprocessing:
        """Transforms raw input data into a collection of datasets,
        with varying preprocessings.

        The helper function initializes an RNG. This RNG is passed to the
        `DatasetCollectionWithPreprocessing` class. When an item (dataset)
        is retrieved, the collection's preprocessing routine uses this stored
        RNG to generate seeds for its individual workers/pipelines, ensuring
        reproducible stochastic transformations from a fixed initial state.

        Args:
            X_raw: single or list of input dataset features, in case of single it
            is converted to list inside get_preprocessed_datasets_helper()
            y_raw: single or list of input dataset labels, in case of single it
            is converted to list inside get_preprocessed_datasets_helper()
            split_fn: A function to dissect a dataset into train and test partition.
            max_data_size: Maximum allowed number of samples within one dataset.
            If None, datasets are not splitted.
            equal_split_size: If True, splits data into equally sized chunks under
            max_data_size.
            If False, splits into chunks of size `max_data_size`, with
            the last chunk having the remainder samples but is dropped if its
            size is less than 2.
        """
        return get_preprocessed_datasets_helper(
            self,
            X_raw,
            y_raw,
            split_fn,
            max_data_size,
            model_type="regressor",
            equal_split_size=equal_split_size,
        )

    def _initialize_model_variables(self) -> tuple[int, np.random.Generator]:
        """Initializes the model, returning byte_size and RNG object."""
        return initialize_model_variables_helper(self, "regressor")

    def _initialize_dataset_preprocessing(
        self, X: XType, y: YType, rng: np.random.Generator
    ) -> tuple[list[RegressorEnsembleConfig], XType, YType, FullSupportBarDistribution]:
        """Prepare ensemble configs and validate X, y for one dataset/chunk.
        Handle the preprocessing of the input (X and y). We also return the
        BarDistribution here, since it is vital for computing the standardized
        target variable in the DatasetCollectionWithPreprocessing class.
        Sets self.inferred_categorical_indices_.
        """
        if self.differentiable_input:
            raise ValueError(
                "Differentiable input is not supported for regressors yet."
            )

        X, y, feature_names_in, n_features_in = validate_Xy_fit(
            X,
            y,
            estimator=self,
            ensure_y_numeric=True,
            max_num_samples=self.inference_config_.MAX_NUMBER_OF_SAMPLES,
            max_num_features=self.inference_config_.MAX_NUMBER_OF_FEATURES,
            ignore_pretraining_limits=self.ignore_pretraining_limits,
        )

        assert isinstance(X, np.ndarray)
        check_cpu_warning(
            self.devices_, X, allow_cpu_override=self.ignore_pretraining_limits
        )

        if feature_names_in is not None:
            self.feature_names_in_ = feature_names_in
        self.n_features_in_ = n_features_in

        self.inferred_categorical_indices_ = infer_categorical_features(
            X=X,
            provided=self.categorical_features_indices,
            min_samples_for_inference=self.inference_config_.MIN_NUMBER_SAMPLES_FOR_CATEGORICAL_INFERENCE,
            max_unique_for_category=self.inference_config_.MAX_UNIQUE_FOR_CATEGORICAL_FEATURES,
            min_unique_for_numerical=self.inference_config_.MIN_UNIQUE_FOR_NUMERICAL_FEATURES,
        )

        # Will convert inferred categorical indices to category dtype,
        # to be picked up by the ord_encoder, as well
        # as handle `np.object` arrays or otherwise `object` dtype pandas columns.
        X = fix_dtypes(X, cat_indices=self.inferred_categorical_indices_)
        # Ensure categories are ordinally encoded
        ord_encoder = get_ordinal_encoder()
        X = process_text_na_dataframe(
            X,
            ord_encoder=ord_encoder,
            fit_encoder=True,  # type: ignore
        )
        self.preprocessor_ = ord_encoder

        possible_target_transforms = get_all_reshape_feature_distribution_preprocessors(
            num_examples=y.shape[0],  # Use length of validated y
            random_state=rng,  # Use the provided rng
        )
        target_preprocessors: list[TransformerMixin | Pipeline | None] = []
        for (
            y_target_preprocessor
        ) in self.inference_config_.REGRESSION_Y_PREPROCESS_TRANSFORMS:
            if y_target_preprocessor is not None:
                preprocessor = possible_target_transforms[y_target_preprocessor]
            else:
                preprocessor = None
            target_preprocessors.append(preprocessor)

        ensemble_configs = EnsembleConfig.generate_for_regression(
            num_estimators=self.n_estimators,
            subsample_size=self.inference_config_.SUBSAMPLE_SAMPLES,
            add_fingerprint_feature=self.inference_config_.FINGERPRINT_FEATURE,
            feature_shift_decoder=self.inference_config_.FEATURE_SHIFT_METHOD,
            polynomial_features=self.inference_config_.POLYNOMIAL_FEATURES,
            max_index=len(X),
            preprocessor_configs=self.inference_config_.PREPROCESS_TRANSFORMS,
            target_transforms=target_preprocessors,
            random_state=rng,
            num_models=len(self.models_),
        )

        self.znorm_space_bardist_ = self.znorm_space_bardist_.to(self.devices_[0])

        assert len(ensemble_configs) == self.n_estimators

        return ensemble_configs, X, y, self.znorm_space_bardist_

    @track_model_call("fit", param_names=["X_preprocessed", "y_preprocessed"])
    def fit_from_preprocessed(
        self,
        X_preprocessed: list[torch.Tensor],
        y_preprocessed: list[torch.Tensor],  # These y are standardized
        cat_ix: list[list[int]],
        configs: list[list[EnsembleConfig]],  # Should be RegressorEnsembleConfig
        *,
        no_refit: bool = True,
    ) -> TabPFNRegressor:
        """Used in Fine-Tuning. Fit the model to preprocessed inputs from torch
        dataloader inside a training loop a Dataset provided by
        get_preprocessed_datasets. This function sets the fit_mode attribute
        to "batched" internally.

        Args:
            X_preprocessed: The input features obtained from the preprocessed Dataset
                The list contains one item for each ensemble predictor.
                use tabpfn.utils.collate_for_tabpfn_dataset to use this function with
                batch sizes of more than one dataset (see examples/tabpfn_finetune.py)
            y_preprocessed: The target variable obtained from the preprocessed Dataset
            cat_ix: categorical indices obtained from the preprocessed Dataset
            configs: Ensemble configurations obtained from the preprocessed Dataset
            no_refit: if True, the classifier will not be reinitialized when calling
                fit multiple times.
        """
        if self.fit_mode != "batched":
            logging.warning(
                "The model was not in 'batched' mode. "
                "Automatically switching to 'batched' mode for finetuning."
            )
            self.fit_mode = "batched"

        # If there is a model, and we are lazy, we skip reinitialization
        if not hasattr(self, "models_") or not no_refit:
            byte_size, rng = self._initialize_model_variables()
        else:
            _, _, byte_size = determine_precision(
                self.inference_precision, self.devices_
            )
            rng = None

        # Create the inference engine
        self.executor_ = create_inference_engine(
            X_train=X_preprocessed,
            y_train=y_preprocessed,
            models=self.models_,
            ensemble_configs=configs,
            cat_ix=cat_ix,
            fit_mode="batched",
            devices_=self.devices_,
            rng=rng,
            n_preprocessing_jobs=self.n_preprocessing_jobs,
            byte_size=byte_size,
            forced_inference_dtype_=self.forced_inference_dtype_,
            memory_saving_mode=self.memory_saving_mode,
            use_autocast_=self.use_autocast_,
            inference_mode=not self.differentiable_input,  # False if differentiable
            # needed (prompt tune)
        )

        return self

    @config_context(transform_output="default")  # type: ignore
    @track_model_call(model_method="fit", param_names=["X", "y"])
    def fit(self, X: XType, y: YType) -> Self:
        """Fit the model.

        Args:
            X: The input data.
            y: The target variable.

        Returns:
            self
        """
        ensemble_configs: list[RegressorEnsembleConfig]

        if self.fit_mode == "batched":
            logging.warning(
                "The model was in 'batched' mode, likely after finetuning. "
                "Automatically switching to 'fit_preprocessors' mode for standard "
                "prediction. The model will be re-initialized."
            )
            self.fit_mode = "fit_preprocessors"

        if not hasattr(self, "models_") or not self.differentiable_input:
            byte_size, rng = self._initialize_model_variables()
            ensemble_configs, X, y, self.znorm_space_bardist_ = (
                self._initialize_dataset_preprocessing(X, y, rng)
            )
        else:  # already fitted and prompt_tuning mode: no cat. features
            _, rng = infer_random_state(self.random_state)
            _, _, byte_size = determine_precision(
                self.inference_precision, self.devices_
            )

        assert len(ensemble_configs) == self.n_estimators

        self.is_constant_target_ = np.unique(y).size == 1
        self.constant_value_ = y[0] if self.is_constant_target_ else None

        if self.is_constant_target_:
            # Use relative epsilon, s.t. it works for small and large constant values
            border_adjustment = max(
                abs(self.constant_value_ * REGRESSION_CONSTANT_TARGET_BORDER_EPSILON),
                REGRESSION_CONSTANT_TARGET_BORDER_EPSILON,
            )

            self.znorm_space_bardist_ = FullSupportBarDistribution(
                borders=torch.tensor(
                    [
                        self.constant_value_ - border_adjustment,
                        self.constant_value_ + border_adjustment,
                    ]
                )
            )
            # No need to create an inference engine for a constant prediction
            return self

        mean, std = np.mean(y), np.std(y)
        self.y_train_std_ = std.item() + 1e-20
        self.y_train_mean_ = mean.item()
        y = (y - self.y_train_mean_) / self.y_train_std_
        self.raw_space_bardist_ = FullSupportBarDistribution(
            self.znorm_space_bardist_.borders * self.y_train_std_ + self.y_train_mean_,
        ).float()

        # Create the inference engine
        self.executor_ = create_inference_engine(
            X_train=X,
            y_train=y,
            models=self.models_,
            ensemble_configs=ensemble_configs,
            cat_ix=self.inferred_categorical_indices_,
            fit_mode=self.fit_mode,
            devices_=self.devices_,
            rng=rng,
            n_preprocessing_jobs=self.n_preprocessing_jobs,
            byte_size=byte_size,
            forced_inference_dtype_=self.forced_inference_dtype_,
            memory_saving_mode=self.memory_saving_mode,
            use_autocast_=self.use_autocast_,
            # TODO: Standard fit usually uses inference_mode=True, before it was enabled
        )

        return self

    @overload
    def predict(
        self,
        X: XType,
        *,
        output_type: Literal["mean", "median", "mode"] = "mean",
        quantiles: list[float] | None = None,
    ) -> np.ndarray: ...

    @overload
    def predict(
        self,
        X: XType,
        *,
        output_type: Literal["quantiles"],
        quantiles: list[float] | None = None,
    ) -> list[np.ndarray]: ...

    @overload
    def predict(
        self,
        X: XType,
        *,
        output_type: Literal["main"],
        quantiles: list[float] | None = None,
    ) -> MainOutputDict: ...

    @overload
    def predict(
        self,
        X: XType,
        *,
        output_type: Literal["full"],
        quantiles: list[float] | None = None,
    ) -> FullOutputDict: ...

    @config_context(transform_output="default")  # type: ignore
    @track_model_call(model_method="predict", param_names=["X"])
    def predict(
        self,
        X: XType,
        *,
        # TODO: support "ei", "pi"
        output_type: OutputType = "mean",
        quantiles: list[float] | None = None,
    ) -> RegressionResultType:
        """Runs the forward() method and then transform the logits
        from the binning space in order to predict target variable.

        Args:
            X: The input data.
            output_type:
                Determines the type of output to return.

                - If `"mean"`, we return the mean over the predicted distribution.
                - If `"median"`, we return the median over the predicted distribution.
                - If `"mode"`, we return the mode over the predicted distribution.
                - If `"quantiles"`, we return the quantiles of the predicted
                    distribution. The parameter `quantiles` determines which
                    quantiles are returned.
                - If `"main"`, we return the all output types above in a dict.
                - If `"full"`, we return the full output of the model, including the
                  logits and the criterion, and all the output types from "main".

            quantiles:
                The quantiles to return if `output="quantiles"`.

                By default, the `[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]`
                quantiles are returned. The predictions per quantile match
                the input order.

        Returns:
            The prediction, which can be a numpy array, a list of arrays (for
            quantiles), or a dictionary with detailed outputs.
        """
        check_is_fitted(self)

        # TODO: Move these at some point to InferenceEngine
        X = validate_X_predict(X, self)

        check_is_fitted(self)

        if quantiles is None:
            quantiles = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        else:
            assert all((0 <= q <= 1) and (isinstance(q, float)) for q in quantiles), (
                "All quantiles must be between 0 and 1 and floats."
            )
        if output_type not in _USABLE_OUTPUT_TYPES:
            raise ValueError(f"Invalid output type: {output_type}")

        if hasattr(self, "is_constant_target_") and self.is_constant_target_:
            return self._handle_constant_target(X.shape[0], output_type, quantiles)

        X = fix_dtypes(X, cat_indices=self.inferred_categorical_indices_)
        X = process_text_na_dataframe(X, ord_encoder=self.preprocessor_)  # type: ignore

        # Runs over iteration engine
        (
            _,
            outputs,  # list of tensors [N_est, N_samples, N_borders] (after forward)
            borders,  # list of numpy arrays containing borders for each estimator
        ) = self.forward(X, use_inference_mode=True)

        # --- Translate probs, average, get final logits ---
        transformed_logits = [
            translate_probs_across_borders(
                logits,
                frm=torch.as_tensor(borders_t, device=logits.device),
                to=self.znorm_space_bardist_.borders.to(logits.device),
            )
            for logits, borders_t in zip(outputs, borders)
        ]
        stacked_logits = torch.stack(transformed_logits, dim=0)
        if self.average_before_softmax:
            logits = stacked_logits.log().mean(dim=0).softmax(dim=-1)
        else:
            logits = stacked_logits.mean(dim=0)

        # Post-process the logits
        logits = logits.log()
        if logits.dtype == torch.float16:
            logits = logits.float()

        # Determine and return intended output type
        logit_to_output = partial(
            _logits_to_output,
            logits=logits,
            criterion=self.raw_space_bardist_,
            quantiles=quantiles,
        )
        if output_type in ["full", "main"]:
            # Create a dictionary of outputs with proper typing via TypedDict
            # Get individual outputs with proper typing
            mean_out = typing.cast("np.ndarray", logit_to_output(output_type="mean"))
            median_out = typing.cast(
                "np.ndarray", logit_to_output(output_type="median")
            )
            mode_out = typing.cast("np.ndarray", logit_to_output(output_type="mode"))
            quantiles_out = typing.cast(
                "list[np.ndarray]",
                logit_to_output(output_type="quantiles"),
            )

            # Create our typed dictionary
            main_outputs = MainOutputDict(
                mean=mean_out,
                median=median_out,
                mode=mode_out,
                quantiles=quantiles_out,
            )

            if output_type == "full":
                # Return full output with criterion and logits
                return FullOutputDict(
                    **main_outputs,
                    criterion=self.raw_space_bardist_,
                    logits=logits,
                )

            return main_outputs

        return logit_to_output(output_type=output_type)

    def forward(
        self,
        X: list[torch.Tensor] | XType,
        *,
        use_inference_mode: bool = False,
    ) -> tuple[torch.Tensor | None, list[torch.Tensor], list[np.ndarray]]:
        """Forward pass for TabPFNRegressor Inference Engine.
        Used in fine-tuning and prediction. Called directly
        in FineTuning training loop or by predict() function
        with the use_inference_mode flag explicitly set to True.

        Iterates over outputs of InferenceEngine.

        Args:
            X: list[torch.Tensor] in fine-tuning, XType in normal predictions.
            use_inference_mode: Flag for inference mode., default at False since
            it is called within predict. During FineTuning forward() is called
            directly by user, so default should be False here.

        Returns:
            A tuple containing:
                - Averaged logits over the ensemble (for fine-tuning).
                - Raw outputs from each estimator in the ensemble.
                - Borders used for each estimator.
        """
        # Scenario 1: Standard inference path
        is_standard_inference = use_inference_mode and not isinstance(
            self.executor_, InferenceEngineBatchedNoPreprocessing
        )

        # Scenario 2: Batched path, typically for fine-tuning with gradients
        is_batched_for_grads = (
            not use_inference_mode
            and isinstance(self.executor_, InferenceEngineBatchedNoPreprocessing)
            and isinstance(X, list)
            and (not X or isinstance(X[0], torch.Tensor))
        )

        assert is_standard_inference or is_batched_for_grads, (
            "Invalid forward pass: Bad combination of inference mode, input X, "
            "or executor type. Ensure call is from standard predict or a "
            "batched fine-tuning context."
        )

        # Specific check for float64 incompatibility if the batched engine is being
        # used, now framed as an assertion that the problematic condition is NOT met.
        assert not (
            isinstance(self.executor_, InferenceEngineBatchedNoPreprocessing)
            and self.forced_inference_dtype_ == torch.float64
        ), (
            "Batched engine error: float64 precision is not supported for the "
            "fine-tuning workflow (requires float32 for backpropagation)."
        )

        # Ensure torch.inference_mode is OFF to allow gradients
        if self.fit_mode in ["fit_preprocessors", "batched"]:
            # only these two modes support this option
            self.executor_.use_torch_inference_mode(use_inference=use_inference_mode)

        check_is_fitted(self)

        std_borders = self.znorm_space_bardist_.borders.cpu().numpy()
        outputs: list[torch.Tensor] = []
        borders: list[np.ndarray] = []

        # Iterate over estimators
        for output, config in self.executor_.iter_outputs(
            X,
            devices=self.devices_,
            autocast=self.use_autocast_,
        ):
            output = output.float()  # noqa: PLW2901
            if self.softmax_temperature != 1:
                output = output / self.softmax_temperature  # noqa: PLW2901

            # BSz.= 1 Scenario, the same as normal predict() function
            # Handled by first if-statement
            config_for_ensemble = config
            if isinstance(config, list) and len(config) == 1:
                single_config = config[0]
                config_for_ensemble = single_config

            if isinstance(config_for_ensemble, RegressorEnsembleConfig):
                borders_t: np.ndarray
                logit_cancel_mask: np.ndarray | None
                descending_borders: bool

                # TODO(eddiebergman): Maybe this could be parallelized or done in fit
                # but I somehow doubt it takes much time to be worth it.
                # One reason to make it worth it is if you want fast predictions, i.e.
                # don't re-do this each time.
                # However it gets a bit more difficult as you need to line up the
                # outputs from `iter_outputs` above (which may be in arbitrary order),
                # along with the specific config the output belongs to. This is because
                # the transformation done to the borders for a given output is dependant
                # upon the target_transform of the config.
                if config_for_ensemble.target_transform is None:
                    borders_t = std_borders.copy()
                    logit_cancel_mask = None
                    descending_borders = False
                else:
                    logit_cancel_mask, descending_borders, borders_t = (
                        transform_borders_one(
                            std_borders,
                            target_transform=config_for_ensemble.target_transform,
                            repair_nan_borders_after_transform=self.inference_config_.FIX_NAN_BORDERS_AFTER_TARGET_TRANSFORM,
                        )
                    )
                    if descending_borders:
                        borders_t = borders_t.flip(-1)  # type: ignore

                borders.append(borders_t)

                if logit_cancel_mask is not None:
                    output = output.clone()  # noqa: PLW2901
                    output[..., logit_cancel_mask] = float("-inf")

            else:
                raise ValueError(
                    "Unexpected config format "
                    "and Batch prediction is not supported yet!"
                )

            outputs.append(output)  # type: ignore

        averaged_logits = None
        all_logits = None
        if outputs:
            all_logits = torch.stack(outputs, dim=0)  # [N_est, N_sampls, N_bord]
            averaged_logits_over_ensemble = torch.mean(all_logits, dim=0)
            averaged_logits = averaged_logits_over_ensemble.transpose(0, 1)

        return averaged_logits, outputs, borders

    def _handle_constant_target(
        self, n_samples: int, output_type: OutputType, quantiles: list[float]
    ) -> RegressionResultType:
        """Handles prediction when the training target `y` was a constant value."""
        constant_prediction = np.full(n_samples, self.constant_value_)
        if output_type in _OUTPUT_TYPES_BASIC:
            return constant_prediction
        if output_type == "quantiles":
            return [np.copy(constant_prediction) for _ in quantiles]

        # Handle "main" and "full"
        main_outputs = MainOutputDict(
            mean=constant_prediction,
            median=np.copy(constant_prediction),
            mode=np.copy(constant_prediction),
            quantiles=[np.copy(constant_prediction) for _ in quantiles],
        )
        if output_type == "full":
            return FullOutputDict(
                **main_outputs,
                criterion=self.znorm_space_bardist_,
                logits=torch.zeros((n_samples, 1)),
            )
        return main_outputs

    def get_embeddings(
        self,
        X: XType,
        data_source: Literal["train", "test"] = "test",
    ) -> np.ndarray:
        """Gets the embeddings for the input data `X`.

        Args:
            X : XType
                The input data.
            data_source : {"train", "test"}, default="test"
                Select the transformer output to return. Use ``"train"`` to obtain
                embeddings from the training tokens and ``"test"`` for the test
                tokens. When ``n_estimators > 1`` the returned array has shape
                ``(n_estimators, n_samples, embedding_dim)``.

        Returns:
            np.ndarray
                The computed embeddings for each fitted estimator.
        """
        return get_embeddings(self, X, data_source)

    def save_fit_state(self, path: Path | str) -> None:
        """Save a fitted regressor, light wrapper around save_fitted_tabpfn_model."""
        save_fitted_tabpfn_model(self, path)

    @classmethod
    def load_from_fit_state(
        cls, path: Path | str, *, device: str | torch.device = "cpu"
    ) -> TabPFNRegressor:
        """Restore a fitted regressor, light wrapper around load_fitted_tabpfn_model."""
        est = load_fitted_tabpfn_model(path, device=device)
        if not isinstance(est, cls):
            raise TypeError(
                f"Attempting to load a '{est.__class__.__name__}' as '{cls.__name__}'"
            )
        return est


def _logits_to_output(
    *,
    output_type: str,
    logits: torch.Tensor,
    criterion: FullSupportBarDistribution,
    quantiles: list[float],
) -> np.ndarray | list[np.ndarray]:
    """Converts raw model logits to the desired prediction format."""
    if output_type == "quantiles":
        return [criterion.icdf(logits, q).cpu().detach().numpy() for q in quantiles]

    # TODO: support
    #   "pi": criterion.pi(logits, np.max(self.y)),
    #   "ei": criterion.ei(logits),
    if output_type == "mean":
        output = criterion.mean(logits)
    elif output_type == "median":
        output = criterion.median(logits)
    elif output_type == "mode":
        output = criterion.mode(logits)
    else:
        raise ValueError(f"Invalid output type: {output_type}")

    return output.cpu().detach().numpy()
